from snowddl.blueprint import BusinessRoleBlueprint, UserBlueprint, WarehouseBlueprint
from snowddl.validator.abc_validator import AbstractValidator


class UserValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(UserBlueprint)

    def validate_blueprint(self, bp: UserBlueprint):
        self._validate_business_roles(bp)
        self._validate_default_warehouse(bp)
        self._validate_user_type_properties(bp)

    def _validate_business_roles(self, bp: UserBlueprint):
        for business_role in bp.business_roles:
            if business_role not in self.config.get_blueprints_by_type(BusinessRoleBlueprint):
                raise ValueError(f"User [{bp.full_name}] references to undefined business role [{business_role}]")

    def _validate_default_warehouse(self, bp: UserBlueprint):
        if bp.default_warehouse and bp.default_warehouse not in self.config.get_blueprints_by_type(WarehouseBlueprint):
            raise ValueError(f"User [{bp.full_name}] references to undefined default warehouse [{bp.default_warehouse}]")

    def _validate_user_type_properties(self, bp: UserBlueprint):
        if bp.type in ("SERVICE", "LEGACY_SERVICE"):
            if bp.first_name:
                raise ValueError(f"Property [first_name] is not allowed for user [{bp.full_name}] with type [{bp.type}]")

            if bp.last_name:
                raise ValueError(f"Property [last_name] is not allowed for user [{bp.full_name}] with type [{bp.type}]")

        if bp.type == "SERVICE":
            if bp.password:
                raise ValueError(f"Property [password] is not allowed for user [{bp.full_name}] with type [{bp.type}]")



from json import loads
from pathlib import Path

from snowddl.blueprint import ObjectType, DataType, BaseDataType
from snowddl.converter.abc_converter import ConvertResult
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter
from snowddl.converter._yaml import YamlLiteralStr, YamlIncludeStr
from snowddl.parser.function import function_json_schema
from snowddl.resolver._utils import dtypes_from_arguments


class FunctionConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FUNCTION

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW USER FUNCTIONS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Skip external functions
            if r["is_external_function"] == "Y":
                continue

            full_name = f"{r['catalog_name']}.{r['schema_name']}.{r['name']}({dtypes_from_arguments(r['arguments'])})"

            existing_objects[full_name] = {
                "database": r["catalog_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "language": r["language"],
                "arguments": r["arguments"],
                "comment": r["description"] if r["description"] else None,
                "is_table_function": r["is_table_function"] == "Y",
                "is_secure": r["is_secure"] == "Y",
                "is_memoizable": r["is_memoizable"] == "Y",
            }

        return existing_objects

    def dump_object(self, row):
        dtypes = dtypes_from_arguments(row["arguments"])

        cur = self.engine.execute_meta(
            "DESC FUNCTION {database:i}.{schema:i}.{name:i}({dtypes:r})",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
                "dtypes": dtypes,
            },
        )

        desc_func_row = {r["property"]: r["value"] for r in cur}

        object_path = (
            self.base_path / self._normalise_name_with_prefix(row["database"]) / self._normalise_name(row["schema"]) / "function"
        )

        data = {
            "language": row["language"],
            "runtime_version": desc_func_row.get("runtime_version"),
            "arguments": self._get_arguments(desc_func_row),
            "returns": self._get_returns(desc_func_row),
            "is_secure": True if row["is_secure"] else None,
            "is_strict": True if desc_func_row.get("null handling") == "RETURNS NULL ON NULL INPUT" else None,
            "is_immutable": True if desc_func_row.get("volatility") == "IMMUTABLE" else None,
            "is_memoizable": True if row["is_memoizable"] else None,
            "imports": self._get_imports(desc_func_row),
            "packages": self._get_packages(desc_func_row),
            "handler": desc_func_row.get("handler"),
            "body": self._get_body_or_include(object_path, row["name"], dtypes, desc_func_row),
            "comment": row.get("comment"),
        }

        if data:
            file_name = self._normalise_name(f"{row['name']}({dtypes}).yaml")
            self._dump_file(object_path / file_name, data, function_json_schema)
            return ConvertResult.DUMP

        return ConvertResult.EMPTY

    def _get_body_or_include(self, object_path: Path, name: str, dtypes: str, desc_func_row: dict):
        if not desc_func_row.get("body"):
            return None

        body = str(desc_func_row["body"])
        language = str(desc_func_row["language"]).lower()

        if self.engine.settings.convert_function_body_to_file:
            if language == "python":
                file_ext = "py"
            elif language == "javascript":
                file_ext = "js"
            else:
                file_ext = language

            file_name = self._normalise_name(f"{name}({dtypes}).{file_ext}")
            self._dump_code(object_path / language / file_name, body)

            return YamlIncludeStr(f"{language}/{file_name}")
        else:
            return YamlLiteralStr(body)

    def _get_arguments(self, desc_func_row: dict):
        if desc_func_row["signature"] == "()":
            return None

        arguments = {}

        for part in str(desc_func_row["signature"])[1:-1].split(", "):
            name, base_dtype = part.split(" ")
            arguments[name.lower()] = str(DataType.from_base_type(BaseDataType[base_dtype]))

        return arguments

    def _get_returns(self, desc_func_row: dict):
        if str(desc_func_row["returns"]).startswith("TABLE"):
            return self._get_returns_table(desc_func_row)

        return self._get_returns_single(desc_func_row)

    def _get_returns_single(self, desc_func_row: dict):
        return str(DataType(desc_func_row["returns"]))

    def _get_returns_table(self, desc_func_row: dict):
        returns = {}

        for part in str(desc_func_row["returns"])[7:-1].split(", "):
            name, base_dtype = part.split(" ")
            returns[name.lower()] = str(DataType.from_base_type(BaseDataType[base_dtype]))

        return returns

    def _get_imports(self, desc_func_row):
        if "imports" not in desc_func_row:
            return None

        if desc_func_row["imports"] == "[]":
            return None

        imports = []

        for part in str(desc_func_row["imports"])[1:-1].split(", "):
            stage, path = part.split("/", 2)
            imports.append(
                {
                    "stage": stage[1:].lower(),
                    "path": path,
                }
            )

        return imports

    def _get_packages(self, desc_func_row):
        if "packages" not in desc_func_row:
            return None

        if desc_func_row["packages"] == "[]":
            return None

        # Snowflake uses single quotes for packages meta
        packages = loads(str(desc_func_row["packages"]).replace("'", '"'))

        return packages



from re import compile, IGNORECASE

from snowddl.blueprint import (
    ExternalTableBlueprint,
    ExternalTableColumn,
    PrimaryKeyBlueprint,
    UniqueKeyBlueprint,
    ForeignKeyBlueprint,
    DataType,
    Ident,
    ObjectType,
    SchemaObjectIdent,
    TableConstraintIdent,
    RowAccessPolicyBlueprint,
    RowAccessPolicyReference,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile

col_type_re = compile(r"^(?P<type>[a-z0-9_]+(\((\d+|int|float)(,(\d+))?\))?)" r"(?P<not_null> NOT NULL)?$", IGNORECASE)

# fmt: off
external_table_json_schema = table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "additionalProperties": {
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string"
                    },
                    "expr": {
                        "type": "string"
                    },
                    "comment": {
                        "type": "string"
                    }
                },
                "required": ["type", "expr"],
                "additionalProperties": False
            }
        },
        "location": {
            "type": "object",
            "properties": {
                "stage": {
                    "type": "string"
                },
                "path": {
                    "type": "string"
                },
                "pattern": {
                    "type": "string"
                },
                "file_format": {
                    "type": "string"
                }
            },
            "required": ["stage", "file_format"],
            "additionalProperties": False
        },
        "partition_by": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "partition_type": {
            "type": "string",
            "pattern": "^[A-Za-z_]+$"
        },
        "auto_refresh": {
            "type": "boolean"
        },
        "refresh_on_create": {
            "type": "boolean"
        },
        "aws_sns_topic": {
            "type": "string"
        },
        "table_format": {
            "type": "string",
            "pattern": "^[A-Za-z_]+$"
        },
        "integration": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        },
        "primary_key": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "unique_keys": {
            "type": "array",
            "items": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1,
            }
        },
        "foreign_keys": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "ref_table": {
                        "type": "string"
                    },
                    "ref_columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "additionalProperties": False
            },
            "minItems": 1
        },
        "row_access_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                },
            },
            "required": ["policy_name", "columns"],
            "additionalProperties": False
        },
    },
    "required": ["location"],
    "additionalProperties": False
}
# fmt: on


class ExternalTableParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("external_table", external_table_json_schema, self.process_external_table)

    def process_external_table(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col in f.params["columns"].items():
            m = col_type_re.match(col["type"])

            if not m:
                raise ValueError(
                    f"Incorrect short syntax for column [{col_name}] in table [{f.database}.{f.schema}.{f.name}]: {col['type']}"
                )

            column_blueprints.append(
                ExternalTableColumn(
                    name=Ident(col_name),
                    type=DataType(m.group("type")),
                    expr=col["expr"],
                    not_null=bool(m.group("not_null")),
                    comment=col.get("comment"),
                )
            )

        bp = ExternalTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            columns=column_blueprints if column_blueprints else None,
            partition_by=[Ident(col_name) for col_name in f.params["partition_by"]] if f.params.get("partition_by") else None,
            partition_type=f.params["partition_type"].upper() if f.params.get("partition_type") else None,
            location_stage=build_schema_object_ident(self.env_prefix, f.params["location"]["stage"], f.database, f.schema),
            location_path=f.params["location"].get("path"),
            location_pattern=f.params["location"].get("pattern"),
            file_format=build_schema_object_ident(self.env_prefix, f.params["location"].get("file_format"), f.database, f.schema)
            if f.params["location"].get("file_format")
            else None,
            refresh_on_create=f.params.get("refresh_on_create", False),
            auto_refresh=f.params.get("auto_refresh", False),
            aws_sns_topic=f.params.get("aws_sns_topic"),
            table_format=f.params["table_format"].upper() if f.params.get("table_format") else None,
            integration=Ident(f.params["integration"]) if f.params.get("integration") else None,
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

        # Constraints

        if f.params.get("primary_key"):
            key_bp = PrimaryKeyBlueprint(
                full_name=TableConstraintIdent(
                    self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in f.params["primary_key"]]
                ),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in f.params["primary_key"]],
                comment=None,
            )

            self.config.add_blueprint(key_bp)

        for columns in f.params.get("unique_keys", []):
            key_bp = UniqueKeyBlueprint(
                full_name=TableConstraintIdent(
                    self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in columns]
                ),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in columns],
                comment=None,
            )

            self.config.add_blueprint(key_bp)

        for fk in f.params.get("foreign_keys", []):
            key_bp = ForeignKeyBlueprint(
                full_name=TableConstraintIdent(
                    self.env_prefix, f.database, f.schema, f.name, columns=[Ident(c) for c in fk["columns"]]
                ),
                table_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                columns=[Ident(c) for c in fk["columns"]],
                ref_table_name=build_schema_object_ident(self.env_prefix, fk["ref_table"], f.database, f.schema),
                ref_columns=[Ident(c) for c in fk["ref_columns"]],
                comment=None,
            )

            self.config.add_blueprint(key_bp)

        # Policies

        if f.params.get("row_access_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["row_access_policy"]["policy_name"], f.database, f.schema
            )

            ref = RowAccessPolicyReference(
                object_type=ObjectType.EXTERNAL_TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["row_access_policy"]["columns"]],
            )

            self.config.add_policy_reference(RowAccessPolicyBlueprint, policy_name, ref)



from snowflake.connector.errors import Error


class SnowDDLExecuteError(Exception):
    def __init__(self, snow_exc: Error, sql: str):
        self.snow_exc = snow_exc
        self.sql = sql

    def verbose_message(self):
        params = {
            "message": self.snow_exc.raw_msg,
            "errno": self.snow_exc.errno,
            "sqlstate": self.snow_exc.sqlstate,
            "sfqid": self.snow_exc.sfqid,
            "sql": self.sql,
        }

        pad_length = max(len(x) for x in params)
        res = ""

        for k in params:
            res += f"    {k.ljust(pad_length)}  =>  {params[k]}\n"

        return "(\n" + res + ")"


class SnowDDLUnsupportedError(Exception):
    pass



from typing import Union

from .ident import AbstractIdent, AbstractIdentWithPrefix, DatabaseIdent, SchemaIdent, SchemaObjectIdent
from .ident_pattern import IdentPattern
from .object_type import ObjectType
from ..model import BaseModelWithConfig


class Grant(BaseModelWithConfig):
    privilege: str
    on: ObjectType
    name: AbstractIdent

    def __eq__(self, other):
        if not isinstance(other, Grant):
            raise NotImplementedError

        if self.privilege != other.privilege:
            return False

        if self.on.singular_for_grant != other.on.singular_for_grant:
            return False

        if self.name != other.name:
            return False

        return True


class AccountGrant(BaseModelWithConfig):
    privilege: str


class FutureGrant(BaseModelWithConfig):
    privilege: str
    on_future: ObjectType
    in_parent: ObjectType
    name: Union[DatabaseIdent, SchemaIdent]

    def is_matching_grant(self, grant: Grant):
        if not grant.on.is_future_grant_supported:
            return False

        if self.privilege != grant.privilege:
            return False

        if self.on_future.singular_for_grant != grant.on.singular_for_grant:
            return False

        if self.in_parent == ObjectType.DATABASE:
            if not isinstance(grant.name, (SchemaIdent, SchemaObjectIdent)):
                return False

            if self.name != grant.name.database_full_name:
                return False

        if self.in_parent == ObjectType.SCHEMA:
            if not isinstance(grant.name, SchemaObjectIdent):
                return False

            if self.name != grant.name.schema_full_name:
                return False

        return True


class GrantPattern(BaseModelWithConfig):
    privilege: str
    on: ObjectType
    pattern: IdentPattern

    def is_matching_grant(self, grant: Grant):
        if self.privilege != grant.privilege:
            return False

        if self.on.singular_for_grant != grant.on.singular_for_grant:
            return False

        if not isinstance(grant.name, AbstractIdentWithPrefix):
            return False

        if not self.pattern.is_match_ident(grant.name):
            return False

        return True


class FutureGrantPattern(BaseModelWithConfig):
    privilege: str
    on_future: ObjectType
    in_parent: ObjectType
    pattern: IdentPattern



from snowddl.blueprint import Ident, TaskBlueprint, AccountObjectIdent, SchemaObjectIdent, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
task_json_schema = {
    "type": "object",
    "properties": {
        "body": {
            "type": "string"
        },
        "scheduling_mode": {
            "type": "string"
        },
        "schedule": {
            "type": "string"
        },
        "after": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "finalize": {
            "type": "string",
        },
        "when": {
            "type": "string"
        },
        "warehouse": {
            "type": "string"
        },
        "user_task_managed_initial_warehouse_size": {
            "type": "string"
        },
        "config": {
            "type": "string",
        },
        "allow_overlapping_execution": {
            "type": "boolean"
        },
        "session_params": {
            "type": "object",
            "additionalProperties": {
                "type": ["boolean", "number", "string"]
            }
        },
        "user_task_timeout_ms": {
            "type": "integer"
        },
        "suspend_task_after_num_failures": {
            "type": "integer"
        },
        "error_integration": {
            "type": "string"
        },
        "success_integration": {
            "type": "string"
        },
        "log_level": {
            "type": "string",
        },
        "task_auto_retry_attempts": {
            "type": "integer"
        },
        "user_task_minimum_trigger_interval_in_seconds": {
            "type": "integer"
        },
        "target_completion_interval": {
            "type": "string",
        },
        "serverless_task_min_statement_size": {
            "type": "string",
        },
        "serverless_task_max_statement_size": {
            "type": "string",
        },
        "comment": {
            "type": "string"
        },
    },
    "required": ["body"],
    "additionalProperties": False
}
# fmt: on


class TaskParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("task", task_json_schema, self.process_task)

    def process_task(self, f: ParsedFile):
        after = None
        finalize = None
        depends_on = set()

        if f.params.get("after"):
            after = [build_schema_object_ident(self.env_prefix, t, f.database, f.schema) for t in f.params.get("after")]
            depends_on.update(after)

        if f.params.get("finalize"):
            finalize = build_schema_object_ident(self.env_prefix, f.params.get("finalize"), f.database, f.schema)
            depends_on.add(finalize)

        bp = TaskBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            body=self.normalise_sql_text_param(f.params["body"]),
            scheduling_mode=f.params.get("scheduling_mode"),
            schedule=f.params.get("schedule"),
            after=after,
            finalize=finalize,
            depends_on=depends_on,
            when=f.params.get("when"),
            warehouse=AccountObjectIdent(self.env_prefix, f.params["warehouse"]) if f.params.get("warehouse") else None,
            user_task_managed_initial_warehouse_size=f.params.get("user_task_managed_initial_warehouse_size"),
            config=f.params.get("config"),
            allow_overlapping_execution=f.params.get("allow_overlapping_execution"),
            session_params=self.normalise_params_dict(f.params.get("session_params")),
            user_task_timeout_ms=f.params.get("user_task_timeout_ms"),
            suspend_task_after_num_failures=f.params.get("suspend_task_after_num_failures"),
            error_integration=Ident(f.params.get("error_integration")) if f.params.get("error_integration") else None,
            success_integration=Ident(f.params.get("success_integration")) if f.params.get("success_integration") else None,
            log_level=f.params.get("log_level"),
            task_auto_retry_attempts=f.params.get("task_auto_retry_attempts"),
            user_task_minimum_trigger_interval_in_seconds=f.params.get("user_task_minimum_trigger_interval_in_seconds"),
            target_completion_interval=f.params.get("target_completion_interval"),
            serverless_task_min_statement_size=f.params.get("serverless_task_min_statement_size"),
            serverless_task_max_statement_size=f.params.get("serverless_task_max_statement_size"),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import FileFormatBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class FileFormatResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.FILE_FORMAT

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW FILE FORMATS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "type": r["type"],
                "format_options": r["format_options"],
                "comment": r["comment"],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(FileFormatBlueprint)

    def create_object(self, bp: FileFormatBlueprint):
        common_query = self._build_common_file_format_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append(
            "CREATE FILE FORMAT {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        create_query.append(common_query)

        self.engine.execute_safe_ddl(create_query)

        self.engine.execute_safe_ddl(
            "COMMENT ON FILE FORMAT {full_name:i} IS {comment}",
            {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: FileFormatBlueprint, row: dict):
        common_query = self._build_common_file_format_sql(bp)

        if not common_query.compare_short_hash(row["comment"]):
            alter_query = self.engine.query_builder()

            if bp.type != row["type"]:
                # Format type cannot be changed with ALTER, must re-create the format entirely
                alter_query.append(
                    "CREATE OR REPLACE FILE FORMAT {full_name:i}",
                    {
                        "full_name": bp.full_name,
                    },
                )

                result = ResolveResult.REPLACE
            else:
                # Individual format settings can be changed with ALTER
                alter_query.append(
                    "ALTER FILE FORMAT {full_name:i} SET",
                    {
                        "full_name": bp.full_name,
                    },
                )

                result = ResolveResult.ALTER

            alter_query.append(common_query)

            self.engine.execute_safe_ddl(alter_query)

            self.engine.execute_safe_ddl(
                "COMMENT ON FILE FORMAT {full_name:i} IS {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": common_query.add_short_hash(bp.comment),
                },
            )

            return result

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP FILE FORMAT {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_file_format_sql(self, bp: FileFormatBlueprint):
        query = self.engine.query_builder()

        query.append_nl(
            "TYPE = {type}",
            {
                "type": bp.type,
            },
        )

        for option_name, option_value in bp.format_options.items():
            query.append_nl(
                "{option_name:r} = {option_value:dp}",
                {
                    "option_name": option_name,
                    "option_value": option_value,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query



from snowddl.blueprint import AuthenticationPolicyBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
authentication_policy_json_schema = {
    "type": "object",
    "properties": {
        "authentication_methods": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mfa_authentication_methods": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "mfa_enrollment": {
            "type": "string"
        },
        "client_types": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "security_integrations": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False,
    "required": [
        "authentication_methods",
        "mfa_authentication_methods",
        "mfa_enrollment",
        "client_types",
        "security_integrations",
    ],
}
# fmt: on


class AuthenticationPolicyParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files(
            "authentication_policy", authentication_policy_json_schema, self.process_authentication_policy
        )

    def process_authentication_policy(self, f: ParsedFile):
        # All parameters are required, since Snowflake keeps changing defaults liberally
        # We cannot trust defaults on this object type
        # https://docs.snowflake.com/en/sql-reference/sql/create-authentication-policy
        bp = AuthenticationPolicyBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            authentication_methods=self.normalise_params_list(f.params.get("authentication_methods")),
            mfa_authentication_methods=self.normalise_params_list(f.params.get("mfa_authentication_methods")),
            mfa_enrollment=f.params.get("mfa_enrollment").upper(),
            client_types=self.normalise_params_list(f.params.get("client_types")),
            security_integrations=self.normalise_params_list(f.params.get("security_integrations")),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from pathlib import Path
from yaml import SafeLoader, Node, add_constructor

from snowddl.fernet.wrapper import FernetWrapper


fernet_wrapper = FernetWrapper()


class SnowDDLLoader(SafeLoader):
    pass


def include_constructor(loader: SnowDDLLoader, node: Node):
    yaml_path = Path(loader.name)

    if not yaml_path.is_file():
        raise ValueError(f"Cannot process !include YAML tag in non-file stream with name [{yaml_path}]")

    include_path = yaml_path.parent / str(node.value)

    if not include_path.is_file():
        raise ValueError(f"File [{include_path}] mentioned in !include YAML tag does not exist")

    if not include_path.is_relative_to(yaml_path.parent):
        raise ValueError(
            f"File [{include_path}] mentioned in !include YAML tag is outside of original YAML file directory [{yaml_path.parent}]"
        )

    return include_path.read_text(encoding="utf-8")


def encrypt_constructor(loader: SnowDDLLoader, node: Node):
    yaml_path = Path(loader.name)

    raise ValueError(f"Detected non-encrypted value in file [{yaml_path}]")


def decrypt_constructor(loader: SnowDDLLoader, node: Node):
    return fernet_wrapper.decrypt(str(node.value))


add_constructor("!include", include_constructor, Loader=SnowDDLLoader)
add_constructor("!encrypt", encrypt_constructor, Loader=SnowDDLLoader)
add_constructor("!decrypt", decrypt_constructor, Loader=SnowDDLLoader)



from snowddl.blueprint import UserBlueprint, build_role_ident
from snowddl.error import SnowDDLExecuteError
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import compare_dynamic_param_value


class UserResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.USER

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW USERS LIKE {env_prefix:ls}",
            {
                "env_prefix": self.config.env_prefix,
            },
        )

        for r in cur:
            if r["owner"] != self.engine.context.current_role:
                continue

            existing_objects[r["name"]] = {
                "name": r["name"],
                "login_name": r["login_name"] if r["login_name"] else None,
                "display_name": r["display_name"] if r["display_name"] else None,
                "first_name": r["first_name"] if r["first_name"] else None,
                "last_name": r["last_name"] if r["last_name"] else None,
                "email": r["email"] if r["email"] else None,
                "disabled": r["disabled"] == "true",
                "default_warehouse": r["default_warehouse"] if r["default_warehouse"] else None,
                "default_namespace": r["default_namespace"] if r["default_namespace"] else None,
                "default_role": r["default_role"] if r["default_role"] else None,
                "type": r["type"] if r["type"] else None,
                "has_password": r["has_password"] == "true",
                "has_rsa_public_key": r["has_rsa_public_key"] == "true",
                "has_mfa": r["has_mfa"] == "true",
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(UserBlueprint)

    def create_object(self, bp: UserBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE USER {name:i}",
            {
                "name": bp.full_name,
            },
        )

        # Common properties
        query.append_nl("LOGIN_NAME = {login_name}", {"login_name": bp.login_name})
        query.append_nl("DISPLAY_NAME = {display_name}", {"display_name": bp.display_name})

        if bp.disabled:
            query.append_nl("DISABLED = {disabled:b}", {"disabled": bp.disabled})

        if bp.first_name:
            query.append_nl("FIRST_NAME = {first_name}", {"first_name": bp.first_name})

        if bp.last_name:
            query.append_nl("LAST_NAME = {last_name}", {"last_name": bp.last_name})

        if bp.email:
            query.append_nl("EMAIL = {email}", {"email": bp.email})

        if bp.default_warehouse:
            query.append_nl("DEFAULT_WAREHOUSE = {default_warehouse}", {"default_warehouse": bp.default_warehouse})

        if bp.default_namespace:
            query.append_nl("DEFAULT_NAMESPACE = {default_namespace}", {"default_namespace": bp.default_namespace})

        query.append_nl("DEFAULT_ROLE = {default_role:i}", {"default_role": self._get_user_role_ident(bp)})

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {"comment": bp.comment})

        # Security properties
        if bp.password:
            query.append_nl("PASSWORD = {password}", {"password": bp.password})

        if bp.rsa_public_key:
            query.append_nl("RSA_PUBLIC_KEY = {rsa_public_key}", {"rsa_public_key": bp.rsa_public_key})

        if bp.rsa_public_key_2:
            query.append_nl("RSA_PUBLIC_KEY_2 = {rsa_public_key_2}", {"rsa_public_key_2": bp.rsa_public_key_2})

        # User type
        if bp.type:
            query.append_nl("TYPE = {type}", {"type": bp.type})

        # Object and session parameters
        query.append(self._build_common_parameters(bp))

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl(
            "GRANT ROLE {user_role:i} TO USER {user_name:i}",
            {
                "user_name": bp.full_name,
                "user_role": self._get_user_role_ident(bp),
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: UserBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._compare_properties(bp, row):
            result = ResolveResult.ALTER

        if self._compare_password(bp, row):
            result = ResolveResult.ALTER

        if self._compare_public_keys(bp, row):
            result = ResolveResult.ALTER

        if self._compare_type(bp, row):
            result = ResolveResult.ALTER

        if self._compare_parameters(bp):
            result = ResolveResult.ALTER

        if self._check_user_role_grant(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP USER {name:i}",
            {
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_common_parameters(self, bp: UserBlueprint):
        query = self.engine.query_builder()

        for param_name, param_value in bp.session_params.items():
            query.append_nl(
                "{param_name:r} = {param_value:dp}",
                {
                    "param_name": param_name,
                    "param_value": param_value,
                },
            )

        return query

    def _compare_properties(self, bp: UserBlueprint, row: dict):
        query = self.engine.query_builder()

        query.append(
            "ALTER USER {name:i} SET",
            {
                "name": bp.full_name,
            },
        )

        if bp.login_name != row["login_name"]:
            query.append_nl("LOGIN_NAME = {login_name}", {"login_name": bp.login_name})

        if bp.display_name != row["display_name"]:
            query.append_nl("DISPLAY_NAME = {display_name}", {"display_name": bp.display_name})

        if bp.disabled != row["disabled"]:
            query.append_nl("DISABLED = {disabled:b}", {"disabled": bp.disabled})

        if bp.first_name != row["first_name"]:
            query.append_nl("FIRST_NAME = {first_name}", {"first_name": bp.first_name})

        if bp.last_name != row["last_name"]:
            query.append_nl("LAST_NAME = {last_name}", {"last_name": bp.last_name})

        if bp.email != row["email"]:
            query.append_nl("EMAIL = {email}", {"email": bp.email})

        if bp.default_warehouse != row["default_warehouse"]:
            query.append_nl("DEFAULT_WAREHOUSE = {default_warehouse}", {"default_warehouse": bp.default_warehouse})

        if bp.default_namespace != row["default_namespace"]:
            query.append_nl("DEFAULT_NAMESPACE = {default_namespace}", {"default_namespace": bp.default_namespace})

        if str(self._get_user_role_ident(bp)) != row["default_role"]:
            query.append_nl("DEFAULT_ROLE = {default_role:i}", {"default_role": self._get_user_role_ident(bp)})

        if bp.comment != row["comment"]:
            query.append_nl("COMMENT = {comment}", {"comment": bp.comment})

        if query.fragment_count() > 1:
            self.engine.execute_safe_ddl(query)
            return True

        return False

    def _compare_password(self, bp: UserBlueprint, row: dict):
        if bp.password and (not row["has_password"] or self.engine.settings.refresh_user_passwords):
            try:
                self.engine.execute_safe_ddl(
                    "ALTER USER {name:i} SET PASSWORD = {password}",
                    {
                        "name": bp.full_name,
                        "password": bp.password,
                    },
                )
            except SnowDDLExecuteError as e:
                # Password rejected due to 'PRIOR_USE'
                # Not an error, skip such user
                if e.snow_exc.errno == 3002:
                    return False
                else:
                    raise

            return True

        if not bp.password and row["has_password"]:
            self.engine.execute_safe_ddl(
                "ALTER USER {name:i} UNSET PASSWORD",
                {
                    "name": bp.full_name,
                },
            )

            return True

        return False

    def _compare_public_keys(self, bp: UserBlueprint, row: dict):
        if not bp.rsa_public_key and not bp.rsa_public_key_2 and not row["has_rsa_public_key"]:
            return False

        existing_public_key = None
        existing_public_key_2 = None

        cur = self.engine.execute_meta(
            "DESC USER {name:i}",
            {
                "name": bp.full_name,
            },
        )

        for r in cur:
            if r["property"] == "RSA_PUBLIC_KEY" and r["value"] != "null":
                existing_public_key = r["value"]

            if r["property"] == "RSA_PUBLIC_KEY_2" and r["value"] != "null":
                existing_public_key_2 = r["value"]

        if bp.rsa_public_key == existing_public_key and bp.rsa_public_key_2 == existing_public_key_2:
            return False

        query = self.engine.query_builder()

        query.append(
            "ALTER USER {name:i} SET",
            {
                "name": bp.full_name,
            },
        )

        query.append_nl("RSA_PUBLIC_KEY = {rsa_public_key}", {"rsa_public_key": bp.rsa_public_key})
        query.append_nl("RSA_PUBLIC_KEY_2 = {rsa_public_key_2}", {"rsa_public_key_2": bp.rsa_public_key_2})

        self.engine.execute_safe_ddl(query)

        return True

    def _compare_type(self, bp: UserBlueprint, row: dict):
        # Workaround for: https://docs.snowflake.com/en/release-notes/bcr-bundles/2025_05/bcr-2067
        if bp.type is None and row["type"] == "PERSON":
            return False

        if bp.type != row["type"]:
            # Type is processed separately, since changing TYPE property impacts visibility of other properties
            self.engine.execute_safe_ddl(
                "ALTER USER {name:i} SET TYPE = {type}",
                {
                    "name": bp.full_name,
                    "type": bp.type,
                },
            )

            return True

        return False

    def _compare_parameters(self, bp: UserBlueprint):
        existing_params = self._get_existing_user_parameters(bp)
        query = self.engine.query_builder()

        query.append(
            "ALTER USER {name:i} SET",
            {
                "name": bp.full_name,
            },
        )

        for param_name, param_value in bp.session_params.items():
            if param_name not in existing_params:
                raise ValueError(f"Unknown parameter [{param_name}] for user [{bp.full_name}]")

            if compare_dynamic_param_value(param_value, existing_params[param_name]["value"]):
                continue

            # At least one parameter in blueprint does not match the existing value
            # Refresh all parameters for such user
            query.append(self._build_common_parameters(bp))
            break

        # If parameter was set on USER level and does not exist in blueprint anymore
        # Unset such parameter and reset it to default
        for param_name, p in existing_params.items():
            if p["level"] == "USER" and p["key"] not in bp.session_params:
                # Setting parameter to NULL equals to UNSET
                query.append_nl(
                    "{param_name:r} = NULL",
                    {
                        "param_name": param_name,
                    },
                )

        if query.fragment_count() > 1:
            self.engine.execute_safe_ddl(query)
            return True

        return False

    def _check_user_role_grant(self, bp: UserBlueprint):
        user_role = self._get_user_role_ident(bp)

        cur = self.engine.execute_meta(
            "SHOW GRANTS TO USER {name:i}",
            {
                "name": bp.full_name,
            },
        )

        for r in cur:
            if r["role"] == str(user_role):
                return False

        self.engine.execute_safe_ddl(
            "GRANT ROLE {user_role:i} TO USER {user_name:i}",
            {
                "user_name": bp.full_name,
                "user_role": user_role,
            },
        )

        return True

    def _get_user_role_ident(self, bp: UserBlueprint):
        return build_role_ident(self.config.env_prefix, bp.full_name.name, self.config.USER_ROLE_SUFFIX)

    def _get_existing_user_parameters(self, bp: UserBlueprint):
        existing_params = {}

        cur = self.engine.execute_meta(
            "SHOW PARAMETERS IN USER {name:i}",
            {
                "name": bp.full_name,
            },
        )

        for r in cur:
            # Network policy is managed via POLICY_REFERENCES in NetworkPolicyResolver
            if r["key"] == "NETWORK_POLICY":
                continue

            existing_params[r["key"]] = r

        return existing_params



from snowddl.blueprint import StreamBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class StreamResolver(AbstractSchemaObjectResolver):
    object_type_to_source_type_map = {
        ObjectType.EXTERNAL_TABLE: "External Table",
        ObjectType.EVENT_TABLE: "Table",
        ObjectType.STAGE: "Stage",
        ObjectType.TABLE: "Table",
        ObjectType.VIEW: "View",
    }

    def get_object_type(self) -> ObjectType:
        return ObjectType.STREAM

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW STREAMS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "source_type": r["source_type"],
                "table_name": r["table_name"],
                "type": r["type"],
                "stale": r["stale"] == "true",
                "mode": r["mode"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(StreamBlueprint)

    def create_object(self, bp: StreamBlueprint):
        query = self._build_create_stream_sql(bp)
        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: StreamBlueprint, row: dict):
        replace_reasons = []

        if row["stale"]:
            replace_reasons.append("Stream is marked as stale")

        if self.object_type_to_source_type_map.get(bp.object_type) != row["source_type"]:
            replace_reasons.append(
                f"Source object type [{str(bp.object_type.name)}] in config does not match source_type [{row['source_type']}] in Snowflake"
            )

        if bp.object_name != row["table_name"]:
            if bp.object_type == ObjectType.STAGE and bp.object_name.name == row["table_name"]:
                # Snowflake bug: SHOW STREAMS may return not fully qualified name for stage
                pass
            else:
                replace_reasons.append(
                    f"Source object name [{bp.object_name}] in config does not match table_name [{row['table_name']}] in Snowflake"
                )

        if bp.append_only != ("APPEND_ONLY" in row["mode"]):
            replace_reasons.append(
                f"APPEND_ONLY={str(bp.append_only)} in config does not match mode [{row['mode']}] in Snowflake"
            )

        if bp.insert_only != ("INSERT_ONLY" in row["mode"]):
            replace_reasons.append(
                f"INSERT_ONLY={str(bp.insert_only)} in config does not match mode [{row['mode']}] in Snowflake"
            )

        if replace_reasons:
            query = self._build_create_stream_sql(bp, True)
            query_with_comment = "\n".join(f"-- {r}" for r in replace_reasons) + "\n" + str(query)

            self.engine.execute_unsafe_ddl(query_with_comment)
            return ResolveResult.REPLACE

        if bp.comment != row["comment"]:
            self.engine.execute_safe_ddl(
                "ALTER STREAM {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP STREAM {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_stream_sql(self, bp: StreamBlueprint, is_replace=False):
        query = self.engine.query_builder()
        query.append("CREATE")

        if is_replace:
            query.append("OR REPLACE")

        query.append(
            "STREAM {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        # TODO: uncomment when COPY GRANTS is supported for streams
        # https://docs.snowflake.com/en/sql-reference/sql/create-stream.html
        # query.append_nl("COPY GRANTS")

        query.append_nl(
            "ON {object_type:r} {object_name:i}",
            {
                "object_type": bp.object_type.singular,
                "object_name": bp.object_name,
            },
        )

        if bp.append_only:
            query.append_nl("APPEND_ONLY = TRUE")

        if bp.insert_only:
            query.append_nl("INSERT_ONLY = TRUE")

        if bp.show_initial_rows:
            query.append_nl("SHOW_INITIAL_ROWS = TRUE")

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query



from snowddl.blueprint import (
    GrantPattern,
    FutureGrantPattern,
    AccountGrant,
    TechnicalRoleBlueprint,
    ObjectType,
    IdentPattern,
    build_role_ident,
)
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
technical_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "future_grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "account_grants": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "comment": {
                "type": "string"
            }
        },
        "additionalProperties": False
    }
}
# fmt: on


class TechnicalRoleParser(AbstractParser):
    def load_blueprints(self):
        # Current config name
        self.parse_multi_entity_file("technical_role", technical_role_json_schema, self.process_technical_role)

        # Backwards compatible name
        self.parse_multi_entity_file("tech_role", technical_role_json_schema, self.process_technical_role)

    def process_technical_role(self, technical_role_name, technical_role_params):
        grant_patterns = []
        future_grant_patterns = []
        account_grants = []

        for definition, pattern_list in technical_role_params.get("grants", {}).items():
            on, privileges = definition.upper().split(":")

            for p in privileges.split(","):
                for pattern in pattern_list:
                    grant_patterns.append(GrantPattern(privilege=p, on=ObjectType[on], pattern=IdentPattern(pattern)))

        for definition, pattern_list in technical_role_params.get("future_grants", {}).items():
            on, privileges = definition.upper().split(":")

            for p in privileges.split(","):
                for pattern in pattern_list:
                    future_grant_patterns.append(
                        FutureGrantPattern(
                            privilege=p,
                            on_future=ObjectType[on],
                            in_parent=ObjectType.SCHEMA if ("." in pattern) else ObjectType.DATABASE,
                            pattern=IdentPattern(pattern),
                        )
                    )

        for privilege in technical_role_params.get("account_grants", []):
            account_grants.append(AccountGrant(privilege=privilege))

        bp = TechnicalRoleBlueprint(
            full_name=build_role_ident(self.env_prefix, technical_role_name, self.config.TECHNICAL_ROLE_SUFFIX),
            grant_patterns=grant_patterns,
            future_grant_patterns=future_grant_patterns,
            account_grants=account_grants,
            comment=technical_role_params.get("comment"),
        )

        self.config.add_blueprint(bp)



from pathlib import Path

from snowddl.blueprint import (
    StageBlueprint,
    StageFileBlueprint,
    Ident,
    SchemaObjectIdent,
    StageFileIdent,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
stage_json_schema = {
    "type": "object",
    "properties": {
        "url": {
            "type": "string"
        },
        "storage_integration": {
            "type": "string"
        },
        "encryption": {
            "type": "object",
            "additionalProperties": {
                "type": ["boolean", "number", "string"]
            }
        },
        "directory": {
            "type": "object",
            "additionalProperties": {
                "type": ["boolean", "number", "string"]
            }
        },
        "file_format": {
            "type": "string"
        },
        "copy_options": {
            "type": "object",
            "additionalProperties": {
                "type": ["array", "boolean", "number", "string"]
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class StageParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("stage", stage_json_schema, self.process_stage)

    def process_stage(self, f: ParsedFile):
        stage_files_dir = f.path.with_suffix("")

        if stage_files_dir.is_dir() and f.params.get("url"):
            raise ValueError("External stage cannot have managed stage files")

        # fmt: off
        bp = StageBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            url=f.params.get("url"),
            storage_integration=Ident(f.params["storage_integration"]) if f.params.get("storage_integration") else None,
            encryption=self.normalise_params_dict(f.params.get("encryption")),
            directory=self.normalise_params_dict(f.params.get("directory")),
            file_format=build_schema_object_ident(self.env_prefix, f.params["file_format"], f.database, f.schema) if f.params.get("file_format") else None,
            copy_options=self.normalise_params_dict(f.params.get("copy_options")),
            upload_stage_files=stage_files_dir.is_dir(),
            comment=f.params.get("comment"),
        )
        # fmt: on

        self.config.add_blueprint(bp)

        # Stage files
        if stage_files_dir.is_dir():
            for local_path in stage_files_dir.glob("**/*"):
                if not local_path.is_file():
                    continue

                stage_path = Path("/") / local_path.relative_to(stage_files_dir)

                bp = StageFileBlueprint(
                    full_name=StageFileIdent(self.env_prefix, f.database, f.schema, f.name, path=stage_path),
                    local_path=local_path,
                    stage_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
                    stage_path=stage_path,
                    comment=None,
                )

                self.config.add_blueprint(bp)



from snowddl.blueprint import AccountParameterBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import compare_dynamic_param_value


class AccountParameterResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.ACCOUNT_PARAMETER

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW PARAMETERS FOR ACCOUNT")

        for r in cur:
            # Network policy for ACCOUNT is managed via POLICY_REFERENCES in NetworkPolicyResolver
            if r["key"] == "NETWORK_POLICY":
                continue

            existing_objects[r["key"]] = {
                "key": r["key"],
                "value": r["value"],
                "default": r["default"],
                "level": r["level"],
                "type": r["type"],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(AccountParameterBlueprint)

    def create_object(self, bp: AccountParameterBlueprint):
        raise ValueError(f"Unknown account parameter [{bp.full_name}]")

    def compare_object(self, bp: AccountParameterBlueprint, row: dict):
        if compare_dynamic_param_value(bp.value, row["value"]):
            return ResolveResult.NOCHANGE

        self.engine.execute_unsafe_ddl(
            "ALTER ACCOUNT SET {param_name:r} = {param_value:dp}",
            {
                "param_name": row["key"],
                "param_value": bp.value,
            },
            condition=self.engine.settings.execute_account_params,
        )

        return ResolveResult.ALTER

    def drop_object(self, row: dict):
        if row["value"] != row["default"] and row["level"] == "ACCOUNT":
            self.engine.execute_unsafe_ddl(
                "ALTER ACCOUNT UNSET {param_name:r}",
                {
                    "param_name": row["key"],
                },
                condition=self.engine.settings.execute_account_params,
            )

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def destroy(self):
        pass



from abc import ABC
from typing import List, Optional

from .ident import AbstractIdent, Ident, SchemaObjectIdent
from .object_type import ObjectType
from ..model import BaseModelWithConfig


class AbstractPolicyReference(BaseModelWithConfig, ABC):
    pass


class AggregationPolicyReference(AbstractPolicyReference):
    object_type: ObjectType
    object_name: SchemaObjectIdent
    columns: List[Ident]


class AuthenticationPolicyReference(AbstractPolicyReference):
    object_type: ObjectType
    object_name: Optional[AbstractIdent] = None


class ForeignKeyReference(BaseModelWithConfig):
    columns: List[Ident]
    ref_table_name: SchemaObjectIdent
    ref_columns: List[Ident]


class IndexReference(BaseModelWithConfig):
    columns: List[Ident]
    include: Optional[List[Ident]]


class MaskingPolicyReference(AbstractPolicyReference):
    object_type: ObjectType
    object_name: SchemaObjectIdent
    columns: List[Ident]


class NetworkPolicyReference(AbstractPolicyReference):
    object_type: ObjectType
    object_name: Optional[AbstractIdent] = None


class ProjectionPolicyReference(AbstractPolicyReference):
    object_type: ObjectType
    object_name: SchemaObjectIdent
    column: Ident


class RowAccessPolicyReference(AbstractPolicyReference):
    object_type: ObjectType
    object_name: SchemaObjectIdent
    columns: List[Ident]


class TagReference(BaseModelWithConfig):
    object_type: ObjectType
    object_name: SchemaObjectIdent
    column_name: Optional[Ident]
    tag_value: str



from argparse import ArgumentParser, HelpFormatter
from os import environ, getcwd
from pydantic import BaseModel
from typing import Optional

from snowddl.app.base import BaseApp
from snowddl.blueprint import (
    AbstractBlueprint,
    DatabaseBlueprint,
    DatabaseIdent,
    SchemaIdent,
    SchemaObjectIdent,
    SchemaBlueprint,
    SchemaObjectBlueprint,
)
from snowddl.config import SnowDDLConfig
from snowddl.parser import singledb_parse_sequence
from snowddl.resolver import singledb_resolve_sequence, singledb_destroy_sequence


class SingleDbApp(BaseApp):
    parse_sequence = singledb_parse_sequence
    resolve_sequence = singledb_resolve_sequence
    destroy_sequence = singledb_destroy_sequence

    def __init__(self):
        self.config_db: Optional[DatabaseIdent] = None
        self.target_db: Optional[DatabaseIdent] = None

        super().__init__()

    def init_arguments_parser(self):
        formatter = lambda prog: HelpFormatter(prog, max_help_position=32)

        parser = ArgumentParser(
            prog="snowddl-singledb",
            description="Special SnowDDL mode to process schema objects of single database only",
            formatter_class=formatter,
        )

        # Config
        parser.add_argument(
            "-c",
            help="Path to config directory OR name of bundled test config (default: current directory)",
            metavar="CONFIG_PATH",
            default=getcwd(),
        )

        # Auth
        parser.add_argument(
            "-a",
            help="Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)",
            metavar="ACCOUNT",
            default=environ.get("SNOWFLAKE_ACCOUNT"),
        )
        parser.add_argument(
            "-u",
            help="Snowflake user name (default: SNOWFLAKE_USER env variable)",
            metavar="USER",
            default=environ.get("SNOWFLAKE_USER"),
        )
        parser.add_argument(
            "-p",
            help="Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)",
            metavar="PASSWORD",
            default=environ.get("SNOWFLAKE_PASSWORD"),
        )
        parser.add_argument(
            "-k",
            help="Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)",
            metavar="PRIVATE_KEY",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PATH"),
        )

        # Role & warehouse
        parser.add_argument(
            "-r",
            help="Snowflake active role (default: SNOWFLAKE_ROLE env variable)",
            metavar="ROLE",
            default=environ.get("SNOWFLAKE_ROLE"),
        )
        parser.add_argument(
            "-w",
            help="Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)",
            metavar="WAREHOUSE",
            default=environ.get("SNOWFLAKE_WAREHOUSE"),
        )

        # SingleDb options
        parser.add_argument(
            "--config-db",
            help="Source database name in config (default: detected automatically if only one database is present in config)",
            default=None,
        )
        parser.add_argument(
            "--target-db", help="Target database name in Snowflake account (default: same as --config-db)", default=None
        )

        # Generic options
        parser.add_argument(
            "--authenticator",
            help="Authenticator: snowflake, externalbrowser, oauth, oauth_snowpark (default: SNOWFLAKE_AUTHENTICATOR env variable)",
            default=environ.get("SNOWFLAKE_AUTHENTICATOR", "snowflake"),
        )
        parser.add_argument(
            "--oauth-token",
            help="Oauth access token (default: SNOWFLAKE_OAUTH_TOKEN env variable)",
            default=environ.get("SNOWFLAKE_OAUTH_TOKEN"),
        )
        parser.add_argument(
            "--passphrase",
            help="Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)",
            default=environ.get("SNOWFLAKE_PRIVATE_KEY_PASSPHRASE"),
        )
        parser.add_argument(
            "--env-prefix",
            help="Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)",
            default=environ.get("SNOWFLAKE_ENV_PREFIX"),
        )
        parser.add_argument(
            "--env-prefix-separator",
            help="Custom separator for Env prefix (supported values are: '__', '_', '$')",
            choices=["__", "_", "$"],
            default=environ.get("SNOWFLAKE_ENV_PREFIX_SEPARATOR", "__"),
        )
        parser.add_argument(
            "--max-workers", help="Maximum number of workers to resolve objects in parallel", default=None, type=int
        )

        # Logging
        parser.add_argument(
            "--log-level", help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO"
        )
        # fmt: off
        parser.add_argument(
            "--show-sql", help="Show executed DDL queries", default=False, action="store_true"
        )
        parser.add_argument(
            "--show-timers", help="Show debug timers", default=False, action="store_true"
        )
        # fmt: on
        parser.add_argument(
            "--show-unused-files", help="Show warnings for unused config files", default=False, action="store_true"
        )

        # Placeholders
        parser.add_argument(
            "--placeholder-path", help="Path to config file with environment-specific placeholders", default=None, metavar=""
        )
        parser.add_argument(
            "--placeholder-values", help="Environment-specific placeholder values in JSON format", default=None, metavar=""
        )

        # Object types
        parser.add_argument(
            "--exclude-object-types", help="Comma-separated list of object types NOT to resolve", default=None, metavar=""
        )
        parser.add_argument(
            "--include-object-types",
            help="Comma-separated list of object types TO resolve, all other types are excluded",
            default=None,
            metavar="",
        )

        # Apply even more unsafe changes
        parser.add_argument(
            "--apply-unsafe",
            help="Additionally apply unsafe changes, which may cause loss of data (ALTER, DROP, etc.)",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-replace-table",
            help="Additionally apply REPLACE TABLE when ALTER TABLE is not possible",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-all-policy", help="Additionally apply changes to all types of POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-aggregation-policy",
            help="Additionally apply changes to AGGREGATION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-masking-policy", help="Additionally apply changes to MASKING POLICIES", default=False, action="store_true"
        )
        parser.add_argument(
            "--apply-projection-policy",
            help="Additionally apply changes to PROJECTION POLICIES",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--apply-row-access-policy",
            help="Additionally apply changes to ROW ACCESS POLICIES",
            default=False,
            action="store_true",
        )

        # Refresh state of specific objects
        parser.add_argument(
            "--refresh-stage-encryption",
            help="Additionally refresh stage encryption parameters for existing external stages",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--refresh-secrets",
            help="Additionally refresh secrets",
            default=False,
            action="store_true",
        )

        # Cloning
        parser.add_argument(
            "--clone-table",
            help="Clone all tables from source database (without env_prefix) to destination database (with env_prefix)",
            default=False,
            action="store_true",
        )
        parser.add_argument(
            "--clone-source-env-prefix",
            help="Clone from another environment with different env_prefix",
        )

        # Subparsers
        subparsers = parser.add_subparsers(dest="action")
        subparsers.required = True

        subparsers.add_parser("plan", help="Resolve objects, apply nothing, display suggested changes")
        subparsers.add_parser("apply", help="Resolve objects, apply safe changes, display suggested unsafe changes")
        subparsers.add_parser(
            "destroy", help="Drop objects with specified --env-prefix, use it to reset dev and test environments"
        )
        subparsers.add_parser("validate", help="Validate config only, do not connect to Snowflake")

        return parser

    def init_config(self):
        config = super().init_config()

        # Init Source DB
        if self.args.get("config_db"):
            self.config_db = DatabaseIdent(config.env_prefix, self.args.get("config_db"))

            if str(self.config_db) not in config.get_blueprints_by_type(DatabaseBlueprint):
                raise ValueError(f"Source database [{self.config_db}] does not exist in config")
        else:
            if len(config.get_blueprints_by_type(DatabaseBlueprint)) > 1:
                raise ValueError(
                    "More than one source database exist in config, please choose a specific database using --source-db argument"
                )

            self.config_db = next(iter(config.get_blueprints_by_type(DatabaseBlueprint).values())).full_name

        # Init Target DB
        if self.args.get("target_db"):
            self.target_db = DatabaseIdent(config.env_prefix, self.args.get("target_db"))
        else:
            self.target_db = self.config_db

        return self.convert_config(config)

    def convert_config(self, original_config: SnowDDLConfig):
        singledb_config = SnowDDLConfig(self.env_prefix)

        for bp_dict in original_config.blueprints.values():
            for bp in bp_dict.values():
                if isinstance(bp, DatabaseBlueprint) and bp.full_name == self.config_db:
                    singledb_config.add_blueprint(self.convert_blueprint(bp))

                if isinstance(bp, (SchemaBlueprint, SchemaObjectBlueprint)) and bp.full_name.database_full_name == self.config_db:
                    singledb_config.add_blueprint(self.convert_blueprint(bp))

        return singledb_config

    def convert_blueprint(self, bp: AbstractBlueprint):
        converted_bp = bp.model_copy(deep=True)

        for field_name, field_value in converted_bp:
            self.convert_object_recursive(getattr(converted_bp, field_name))

        return converted_bp

    def convert_object_recursive(self, obj):
        if isinstance(obj, BaseModel):
            for field_name, field_value in obj:
                self.convert_object_recursive(getattr(obj, field_name))

        if isinstance(obj, list):
            for item in obj:
                self.convert_object_recursive(item)

        if isinstance(obj, dict):
            for item in obj.values():
                self.convert_object_recursive(item)

        if isinstance(obj, (DatabaseIdent, SchemaIdent, SchemaObjectIdent)):
            obj.database = self.target_db.database

        return obj

    def init_settings(self):
        settings = super().init_settings()
        settings.include_databases = [self.target_db]
        settings.ignore_ownership = True

        return settings

    def execute(self):
        if self.args.get("action") == "validate":
            return

        error_count = 0

        with self.get_engine() as engine:
            self.output_engine_context(engine)

            if self.args.get("action") == "destroy":
                for resolver_cls in self.destroy_sequence:
                    with self.measure_elapsed_time(resolver_cls.__name__):
                        resolver = resolver_cls(engine)
                        resolver.destroy()

                    error_count += len(resolver.errors)

            else:
                for resolver_cls in self.resolve_sequence:
                    with self.measure_elapsed_time(resolver_cls.__name__):
                        resolver = resolver_cls(engine)
                        resolver.resolve()

                    error_count += len(resolver.errors)

            engine.connection.close()

            self.output_engine_stats(engine)
            self.output_engine_warnings(engine)

            if self.args.get("show_timers"):
                self.output_app_timers()

            if self.args.get("show_sql"):
                self.output_executed_ddl(engine)

            self.output_suggested_ddl(engine)

            if error_count > 0:
                exit(8)


def entry_point():
    app = SingleDbApp()
    app.execute()



from snowddl.blueprint import (
    BusinessRoleBlueprint,
    DatabaseBlueprint,
    SchemaBlueprint,
    TechnicalRoleBlueprint,
    WarehouseBlueprint,
)
from snowddl.validator.abc_validator import AbstractValidator


class BusinessRoleValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(BusinessRoleBlueprint)

    def validate_blueprint(self, bp: BusinessRoleBlueprint):
        self._validate_database_schema_patterns(bp)
        self._validate_warehouse(bp)
        self._validate_technical_roles(bp)

    def _validate_database_schema_patterns(self, bp: BusinessRoleBlueprint):
        for database_name_pattern in bp.database_owner:
            if not self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, database_name_pattern):
                raise ValueError(
                    f"Business role [{bp.full_name}] database_owner pattern [{database_name_pattern}] "
                    f"does not match any databases"
                )

        for database_name_pattern in bp.database_write:
            if not self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, database_name_pattern):
                raise ValueError(
                    f"Business role [{bp.full_name}] database_write pattern [{database_name_pattern}] "
                    f"does not match any databases"
                )

        for database_name_pattern in bp.database_read:
            if not self.config.get_blueprints_by_type_and_pattern(DatabaseBlueprint, database_name_pattern):
                raise ValueError(
                    f"Business role [{bp.full_name}] database_read pattern [{database_name_pattern}] "
                    f"does not match any databases"
                )

        for schema_name_pattern in bp.schema_owner:
            if not self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, schema_name_pattern):
                raise ValueError(
                    f"Business role [{bp.full_name}] schema_owner pattern [{schema_name_pattern}] " f"does not match any schemas"
                )

        for schema_name_pattern in bp.schema_write:
            if not self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, schema_name_pattern):
                raise ValueError(
                    f"Business role [{bp.full_name}] schema_write pattern [{schema_name_pattern}] " f"does not match any schemas"
                )

        for schema_name_pattern in bp.schema_read:
            if not self.config.get_blueprints_by_type_and_pattern(SchemaBlueprint, schema_name_pattern):
                raise ValueError(
                    f"Business role [{bp.full_name}] schema_read pattern [{schema_name_pattern}] " f"does not match any schemas"
                )

    def _validate_warehouse(self, bp: BusinessRoleBlueprint):
        for warehouse_name in bp.warehouse_usage:
            if warehouse_name not in self.config.get_blueprints_by_type(WarehouseBlueprint):
                raise ValueError(
                    f"Business role [{bp.full_name}] warehouse_usage references to " f"undefined warehouse [{warehouse_name}]"
                )

        for warehouse_name in bp.warehouse_monitor:
            if warehouse_name not in self.config.get_blueprints_by_type(WarehouseBlueprint):
                raise ValueError(
                    f"Business role [{bp.full_name}] warehouse_monitor references to " f"undefined warehouse [{warehouse_name}]"
                )

    def _validate_technical_roles(self, bp: BusinessRoleBlueprint):
        for technical_role_name in bp.technical_roles:
            if technical_role_name not in self.config.get_blueprints_by_type(TechnicalRoleBlueprint):
                raise ValueError(
                    f"Business role [{bp.full_name}] references to " f"undefined technical role [{technical_role_name}]"
                )



from snowddl.blueprint import EventTableBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
event_table_json_schema = {
    "type": "object",
    "properties": {
        "change_tracking": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        },
    },
    "additionalProperties": False
}
# fmt: on


class EventTableParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("event_table", event_table_json_schema, self.process_event_table)

    def process_event_table(self, f: ParsedFile):
        bp = EventTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            change_tracking=f.params.get("change_tracking", False),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import (
    DatabaseIdent,
    SchemaIdent,
    ObjectType,
    SnapshotSetBlueprint,
    SchemaObjectIdent,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
snapshot_set_json_schema = {
    "type": "object",
    "properties": {
        "object_type": {
            "type": "string"
        },
        "object_name": {
            "type": "string"
        },
        "snapshot_policy": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class SnapshotSetParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("snapshot_set", snapshot_set_json_schema, self.process_snapshot_set)

    def process_snapshot_set(self, f: ParsedFile):
        object_type = ObjectType[f.params["object_type"]]

        if object_type == ObjectType.DATABASE:
            object_name = DatabaseIdent(self.env_prefix, f.params["object_name"])
        elif object_type == ObjectType.SCHEMA:
            object_name = SchemaIdent(self.env_prefix, *f.params["object_name"].split(".", 2))
        else:
            object_name = build_schema_object_ident(self.env_prefix, f.params["object_name"], f.database, f.schema)

        bp = SnapshotSetBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            object_type=ObjectType[f.params["object_type"]],
            object_name=object_name,
            snapshot_policy=build_schema_object_ident(self.env_prefix, f.params["snapshot_policy"], f.database, f.schema) if f.params.get("snapshot_policy") else None,
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import BusinessRoleBlueprint, RoleBlueprint
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver


class BusinessRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.BUSINESS_ROLE_SUFFIX

    def get_blueprints(self):
        return {
            full_name: self.transform_blueprint(business_role_bp)
            for full_name, business_role_bp in self.config.get_blueprints_by_type(BusinessRoleBlueprint).items()
        }

    def transform_blueprint(self, business_role_bp: BusinessRoleBlueprint) -> RoleBlueprint:
        grants = []

        # Database roles
        for database_name_pattern in business_role_bp.database_owner:
            grants.extend(self.build_database_role_grants(database_name_pattern, self.config.OWNER_ROLE_TYPE))

        for database_name_pattern in business_role_bp.database_write:
            grants.extend(self.build_database_role_grants(database_name_pattern, self.config.WRITE_ROLE_TYPE))

        for database_name_pattern in business_role_bp.database_read:
            grants.extend(self.build_database_role_grants(database_name_pattern, self.config.READ_ROLE_TYPE))

        # Schema roles
        for schema_name_pattern in business_role_bp.schema_owner:
            grants.extend(self.build_schema_role_grants(schema_name_pattern, self.config.OWNER_ROLE_TYPE))

        for schema_name_pattern in business_role_bp.schema_write:
            grants.extend(self.build_schema_role_grants(schema_name_pattern, self.config.WRITE_ROLE_TYPE))

        for schema_name_pattern in business_role_bp.schema_read:
            grants.extend(self.build_schema_role_grants(schema_name_pattern, self.config.READ_ROLE_TYPE))

        # Share roles
        for share_name in business_role_bp.share_read:
            grants.append(self.build_share_read_grant(share_name))

        # Warehouse roles
        for warehouse_name in business_role_bp.warehouse_usage:
            grants.append(self.build_warehouse_role_grant(warehouse_name, self.config.USAGE_ROLE_TYPE))

        for warehouse_name in business_role_bp.warehouse_monitor:
            grants.append(self.build_warehouse_role_grant(warehouse_name, self.config.MONITOR_ROLE_TYPE))

        # Technical roles
        for technical_role_name in business_role_bp.technical_roles:
            grants.append(self.build_technical_role_grant(technical_role_name))

        # Global roles
        for global_role_name in business_role_bp.global_roles:
            grants.append(self.build_global_role_grant(global_role_name))

        return RoleBlueprint(
            full_name=business_role_bp.full_name,
            grants=grants,
            comment=business_role_bp.comment,
        )



from snowddl.blueprint import (
    DatabaseIdent,
    FutureGrant,
    Grant,
    RoleBlueprint,
    SchemaBlueprint,
    build_role_ident,
)
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class SchemaReadRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.SCHEMA_ACCESS_ROLE_SUFFIX

    def get_role_type(self):
        return self.config.READ_ROLE_TYPE

    def get_blueprints(self):
        blueprints = []

        for schema_bp in self.config.get_blueprints_by_type(SchemaBlueprint).values():
            schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

            if schema_permission_model.ruleset.create_schema_read_role:
                blueprints.append(self.get_blueprint_read_role(schema_bp))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_read_role(self, schema_bp: SchemaBlueprint):
        grants = []
        future_grants = []

        schema_permission_model = self.config.get_permission_model(schema_bp.permission_model)

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.DATABASE,
                name=DatabaseIdent(schema_bp.full_name.env_prefix, schema_bp.full_name.database),
            )
        )

        grants.append(
            Grant(
                privilege="USAGE",
                on=ObjectType.SCHEMA,
                name=schema_bp.full_name,
            )
        )

        for model_future_grant in schema_permission_model.read_future_grants:
            future_grants.append(
                FutureGrant(
                    privilege=model_future_grant.privilege,
                    on_future=model_future_grant.on,
                    in_parent=ObjectType.SCHEMA,
                    name=schema_bp.full_name,
                )
            )

        bp = RoleBlueprint(
            full_name=build_role_ident(
                self.config.env_prefix,
                schema_bp.full_name.database,
                schema_bp.full_name.schema,
                self.config.READ_ROLE_TYPE,
                self.get_role_suffix(),
            ),
            grants=grants,
            future_grants=future_grants,
        )

        return bp



from snowddl.blueprint import WarehouseBlueprint, Ident, AccountObjectIdent
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
warehouse_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "type": {
                "type": "string"
            },
            "size": {
                "type": "string"
            },
            "min_cluster_count": {
                "type": "integer"
            },
            "max_cluster_count": {
                "type": "integer"
            },
            "scaling_policy": {
                "type": "string"
            },
            "auto_suspend": {
                "type": "integer"
            },
            "resource_monitor": {
                "type": "string"
            },
            "global_resource_monitor": {
                "type": "string"
            },
            "enable_query_acceleration": {
                "type": "boolean"
            },
            "query_acceleration_max_scale_factor": {
                "type": "integer"
            },
            "warehouse_params": {
                "type": "object",
                "additionalProperties": {
                    "type": ["boolean", "number", "string"]
                }
            },
            "resource_constraint": {
                "type": "string"
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["size"],
        "additionalProperties": False
    }
}
# fmt: on


class WarehouseParser(AbstractParser):
    def load_blueprints(self):
        self.parse_multi_entity_file("warehouse", warehouse_json_schema, self.process_warehouse)

    def process_warehouse(self, warehouse_name, warehouse_params):
        warehouse_type = warehouse_params.get("type", "STANDARD").upper()
        resource_constraint = warehouse_params.get("resource_constraint")
        resource_monitor = None

        if warehouse_params.get("resource_monitor"):
            resource_monitor = AccountObjectIdent(self.env_prefix, warehouse_params["resource_monitor"])

        if warehouse_params.get("global_resource_monitor"):
            resource_monitor = Ident(warehouse_params["global_resource_monitor"])

        bp = WarehouseBlueprint(
            full_name=AccountObjectIdent(self.env_prefix, warehouse_name),
            type=warehouse_type,
            size=warehouse_params["size"],
            auto_suspend=warehouse_params.get("auto_suspend", 60),
            min_cluster_count=warehouse_params.get("min_cluster_count", 1),
            max_cluster_count=warehouse_params.get("max_cluster_count", 1),
            scaling_policy=warehouse_params.get("scaling_policy", "STANDARD").upper(),
            resource_monitor=resource_monitor,
            enable_query_acceleration=warehouse_params.get("enable_query_acceleration", False),
            query_acceleration_max_scale_factor=warehouse_params.get("query_acceleration_max_scale_factor", 8),
            warehouse_params=self.normalise_params_dict(warehouse_params.get("warehouse_params", {})),
            resource_constraint=resource_constraint,
            comment=warehouse_params.get("comment"),
        )

        self.config.add_blueprint(bp)



from typing import Union

from .data_type import BaseDataType
from .ident import (
    AbstractIdent,
    AccountObjectIdent,
    DatabaseIdent,
    DatabaseRoleIdent,
    Ident,
    SchemaIdent,
    SchemaObjectIdent,
    SchemaObjectIdentWithArgs,
)
from .object_type import ObjectType


def build_schema_object_ident(env_prefix, object_name, context_database_name, context_schema_name) -> SchemaObjectIdent:
    # Function or procedure identifier with arguments
    if object_name.endswith(")"):
        object_name, data_types_str = object_name.rstrip(")").split("(")
        data_types = [BaseDataType[dt.strip(" ").upper()] for dt in data_types_str.split(",")] if data_types_str else []
    else:
        data_types = None

    parts = object_name.split(".")
    parts_len = len(parts)

    if parts_len > 3:
        raise ValueError(f"Too many delimiters in schema object identifier [{object_name}]")

    # Add context db_name to "schema_name.object_name" or "object_name" identifier
    if parts_len <= 2:
        parts.insert(0, context_database_name)

    # Add context schema_name to "object_name" identifier
    if parts_len == 1:
        parts.insert(1, context_schema_name)

    # Function or procedure identifier with arguments
    if isinstance(data_types, list):
        return SchemaObjectIdentWithArgs(env_prefix, parts[0], parts[1], parts[2], data_types=data_types)

    return SchemaObjectIdent(env_prefix, parts[0], parts[1], parts[2])


def build_role_ident(env_prefix, *args: Union[AbstractIdent, str]) -> AccountObjectIdent:
    return AccountObjectIdent(env_prefix, f"{'__'.join(str(a) for a in args)}")


def build_grant_name_ident(env_prefix: str, grant_name: str, object_type: ObjectType):
    if grant_name.startswith(env_prefix):
        grant_env_prefix = env_prefix
        grant_name = grant_name.removeprefix(env_prefix)
    else:
        grant_env_prefix = ""

    parts = [p.strip('"') for p in grant_name.split(".")]
    last_part = parts[-1]

    if len(parts) == 3:
        # Extract data types from arguments of FUNCTION or PROCEDURE
        if object_type.is_overloading_supported:
            start_dtypes_idx = last_part.index("(")
            finish_dtypes_idx = last_part.index(")")

            parts[-1] = last_part[0:start_dtypes_idx]
            arguments_str = last_part[start_dtypes_idx + 1 : finish_dtypes_idx]
            data_types = []

            if arguments_str:
                for arg in arguments_str.split(","):
                    data_types.append(BaseDataType[arg.strip(" ").split(" ")[-1]])

            return SchemaObjectIdentWithArgs(grant_env_prefix, parts[0], parts[1], parts[2], data_types=data_types)

        return SchemaObjectIdent(grant_env_prefix, parts[0], parts[1], parts[2])

    if len(parts) == 2:
        if object_type == ObjectType.DATABASE_ROLE:
            return DatabaseRoleIdent(grant_env_prefix, parts[0], parts[1])

        return SchemaIdent(grant_env_prefix, parts[0], parts[1])

    if len(parts) == 1:
        if object_type == ObjectType.DATABASE:
            return DatabaseIdent(grant_env_prefix, parts[0])

        return AccountObjectIdent(grant_env_prefix, parts[0])

    raise ValueError(f"Unexpected grant name format [{grant_name}] in Snowflake for object type [{object_type}]")


def build_future_grant_name_ident(object_type: ObjectType, grant_name: str):
    env_prefix = ""

    parts = [p.strip('"') for p in grant_name.split(".")]
    parts.pop()  # Remove object type component from future grant names

    if len(parts) == 2:
        return SchemaIdent(env_prefix, parts[0], parts[1])

    if len(parts) == 1:
        return DatabaseIdent(env_prefix, parts[0])

    raise ValueError(f"Unexpected grant name format [{grant_name}] in Snowflake for object type [{object_type}]")


def build_default_namespace_ident(env_prefix, default_namespace):
    if "." in default_namespace:
        return SchemaIdent(env_prefix, *default_namespace.split(".", 1))

    return DatabaseIdent(env_prefix, default_namespace)


def build_share_read_ident(share_name: str) -> Union[Ident, DatabaseRoleIdent]:
    if "." in share_name:
        # Shares are global, so database roles inside shares are global as well
        # No env prefix for global objects
        return DatabaseRoleIdent("", *share_name.split(".", 2))
    else:
        return Ident(share_name)



from snowddl.blueprint import ObjectType, RoleBlueprint, SchemaBlueprint, TechnicalRoleBlueprint, Grant, FutureGrant
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver


class TechnicalRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.TECHNICAL_ROLE_SUFFIX

    def get_blueprints(self):
        return {
            full_name: self.transform_blueprint(business_role_bp)
            for full_name, business_role_bp in self.config.get_blueprints_by_type(TechnicalRoleBlueprint).items()
        }

    def transform_blueprint(self, technical_role_bp: TechnicalRoleBlueprint):
        grants = []
        future_grants = []

        for grant_pattern in technical_role_bp.grant_patterns:
            for obj_bp in self.config.get_blueprints_by_type_and_pattern(
                grant_pattern.on.blueprint_cls, grant_pattern.pattern
            ).values():
                grants.append(
                    Grant(
                        privilege=grant_pattern.privilege,
                        on=grant_pattern.on,
                        name=obj_bp.full_name,
                    ),
                )

        for future_grant_pattern in technical_role_bp.future_grant_patterns:
            for obj_bp in self.config.get_blueprints_by_type_and_pattern(
                future_grant_pattern.in_parent.blueprint_cls, future_grant_pattern.pattern
            ).values():
                future_grants.append(
                    FutureGrant(
                        privilege=future_grant_pattern.privilege,
                        on_future=future_grant_pattern.on_future,
                        in_parent=future_grant_pattern.in_parent,
                        name=obj_bp.full_name,
                    ),
                )

                if future_grant_pattern.in_parent != ObjectType.DATABASE:
                    continue

                # Future grants for each known schema on SCHEMA level
                # It is required to alleviate SCHEMA level grants overriding DATABASE level grants
                # https://docs.snowflake.com/en/sql-reference/sql/grant-privilege#future-grants-on-database-or-schema-objects
                for schema_bp in self.config.get_blueprints_by_type(SchemaBlueprint).values():
                    if obj_bp.full_name != schema_bp.full_name.database_full_name:
                        continue

                    future_grants.append(
                        FutureGrant(
                            privilege=future_grant_pattern.privilege,
                            on_future=future_grant_pattern.on_future,
                            in_parent=ObjectType.SCHEMA,
                            name=schema_bp.full_name,
                        )
                    )

        return RoleBlueprint(
            full_name=technical_role_bp.full_name,
            grants=grants,
            future_grants=future_grants,
            account_grants=technical_role_bp.account_grants,
            comment=technical_role_bp.comment,
        )



from functools import partial

from snowddl.blueprint import DatabaseRoleBlueprint, DatabaseRoleIdent, IdentPattern, GrantPattern, ObjectType
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
database_role_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "grants": {
                "type": "object",
                "additionalProperties": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["grants"],
        "additionalProperties": False
    }
}
# fmt: on


class DatabaseRoleParser(AbstractParser):
    def load_blueprints(self):
        for database_name in self.get_database_names():
            self.parse_multi_entity_file(
                f"{database_name}/database_role",
                database_role_json_schema,
                partial(self.process_database_role, database_name=database_name),
            )

    def process_database_role(self, role_name, role_params, database_name):
        bp = DatabaseRoleBlueprint(
            full_name=DatabaseRoleIdent(self.env_prefix, database_name, role_name),
            grant_patterns=self.get_grant_patterns(role_params),
            comment=role_params.get("comment"),
        )

        self.config.add_blueprint(bp)

    def get_grant_patterns(self, role_params):
        grant_patterns = []

        for definition, pattern_list in role_params["grants"].items():
            on, privileges = definition.upper().split(":")

            for p in privileges.split(","):
                for pattern in pattern_list:
                    grant_patterns.append(GrantPattern(privilege=p, on=ObjectType[on], pattern=IdentPattern(pattern)))

        return grant_patterns



from pathlib import Path
from typing import Dict, Optional

from snowddl.blueprint import DatabaseIdent
from snowddl.parser.abc_parser import AbstractParser


# fmt: off
placeholder_json_schema = {
    "type": "object",
    "additionalProperties": {
        "anyOf": [
            {
                "type": ["boolean", "number", "string"]
            },
            {
                "type": "array",
                "items": {
                    "type": ["boolean", "number", "string"]
                },
            },
        ]
    }
}
# fmt: on


class PlaceholderParser(AbstractParser):
    def load_blueprints(self):
        # This is a special parser that does not load any blueprints, but it loads placeholders instead
        pass

    def load_placeholders(
        self, placeholder_path: Optional[Path] = None, placeholder_values: Optional[Dict] = None, args: Optional[Dict] = None
    ):
        # 1) Start with standard placeholders, always available
        placeholders = {
            "ENV_PREFIX": self.env_prefix,
        }

        if args and args.get("target_db"):
            placeholders["TARGET_DB"] = str(DatabaseIdent(self.env_prefix, args.get("target_db")))

        # 2) Merge with placeholders from normal config file
        placeholders.update(self.normalise_params_dict(self.parse_single_entity_file("placeholder", placeholder_json_schema)))

        # 3) Merge with placeholders from override config file
        if placeholder_path:
            placeholders.update(self.normalise_params_dict(self.parse_external_file(placeholder_path, placeholder_json_schema)))

        # 4) Merge with explicit placeholder values
        if placeholder_values:
            placeholders.update(self.normalise_params_dict(placeholder_values))

        for name, value in placeholders.items():
            self.config.add_placeholder(name, value)



from snowddl.blueprint import AlertBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class AlertResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.ALERT

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW ALERTS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "warehouse": r["warehouse"],
                "schedule": r["schedule"],
                "state": r["state"],
                "condition": r["condition"],
                "action": r["action"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(AlertBlueprint)

    def create_object(self, bp: AlertBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE ALERT {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.warehouse:
            query.append_nl(
                "WAREHOUSE = {warehouse:i}",
                {
                    "warehouse": bp.warehouse,
                },
            )

        query.append_nl(
            "SCHEDULE = {schedule}",
            {
                "schedule": bp.schedule,
            },
        )

        query.append_nl("IF(EXISTS(")
        query.append_nl(bp.condition)
        query.append_nl("))")

        query.append_nl("THEN")
        query.append_nl(bp.action)

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: AlertBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.warehouse != row["warehouse"]:
            if bp.warehouse:
                self.engine.execute_safe_ddl(
                    "ALTER ALERT {full_name:i} SET WAREHOUSE = {warehouse:i}",
                    {
                        "full_name": bp.full_name,
                        "warehouse": bp.warehouse,
                    },
                )
            else:
                self.engine.execute_safe_ddl(
                    "ALTER ALERT {full_name:i} UNSET WAREHOUSE",
                    {
                        "full_name": bp.full_name,
                        "warehouse": bp.warehouse,
                    },
                )

            result = ResolveResult.ALTER

        if str(bp.schedule) != row["schedule"]:
            self.engine.execute_safe_ddl(
                "ALTER ALERT {full_name:i} SET SCHEDULE = {schedule}",
                {
                    "full_name": bp.full_name,
                    "schedule": bp.schedule,
                },
            )

            result = ResolveResult.ALTER

        if str(bp.condition) != row["condition"]:
            query = self.engine.query_builder()

            query.append(
                "ALTER ALERT {full_name:i} MODIFY CONDITION EXISTS (",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(bp.condition)
            query.append_nl(")")

            self.engine.execute_safe_ddl(query)

            result = ResolveResult.ALTER

        if str(bp.action) != row["action"]:
            query = self.engine.query_builder()

            query.append(
                "ALTER ALERT {full_name:i} MODIFY ACTION",
                {
                    "full_name": bp.full_name,
                },
            )

            query.append_nl(bp.action)

            self.engine.execute_safe_ddl(query)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "DROP ALERT {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP



from .abc_validator import AbstractValidator
from .business_role import BusinessRoleValidator
from .database import DatabaseValidator
from .dynamic_table import DynamicTableValidator
from .hybrid_table import HybridTableValidator
from .schema import SchemaValidator
from .stream import StreamValidator
from .task import TaskValidator
from .technical_role import TechnicalRoleValidator
from .user import UserValidator
from .view import ViewValidator
from .warehouse import WarehouseValidator


default_validate_sequence = [
    WarehouseValidator,
    DatabaseValidator,
    SchemaValidator,
    DynamicTableValidator,
    HybridTableValidator,
    TaskValidator,
    ViewValidator,
    StreamValidator,
    TechnicalRoleValidator,
    BusinessRoleValidator,
    UserValidator,
]



from snowddl.blueprint import UniqueKeyBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class UniqueKeyResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.UNIQUE_KEY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}
        constraints_by_name = {}

        cur = self.engine.execute_meta(
            "SHOW UNIQUE KEYS IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            # Constraint for Hybrid tables are handled separately
            if r["comment"] == ObjectType.HYBRID_TABLE.name:
                continue

            if r["constraint_name"] not in constraints_by_name:
                constraints_by_name[r["constraint_name"]] = {
                    "database": r["database_name"],
                    "schema": r["schema_name"],
                    "table": r["table_name"],
                    "columns": {r["key_sequence"]: r["column_name"]},
                }
            else:
                constraints_by_name[r["constraint_name"]]["columns"][r["key_sequence"]] = r["column_name"]

        for c in constraints_by_name.values():
            columns_list = [c["columns"][k] for k in sorted(c["columns"])]
            full_name = f"{c['database']}.{c['schema']}.{c['table']}({','.join(columns_list)})"

            existing_objects[full_name] = {
                "database": c["database"],
                "schema": c["schema"],
                "table": c["table"],
                "columns": columns_list,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(UniqueKeyBlueprint)

    def create_object(self, bp: UniqueKeyBlueprint):
        self.engine.execute_safe_ddl(
            "ALTER TABLE {table_name:i} ADD UNIQUE ({columns:i})",
            {
                "table_name": bp.table_name,
                "columns": bp.columns,
            },
        )

        return ResolveResult.CREATE

    def compare_object(self, bp: UniqueKeyBlueprint, row: dict):
        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl(
            "ALTER TABLE {database:i}.{schema:i}.{table:i} DROP UNIQUE ({columns:i})",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table": row["table"],
                "columns": row["columns"],
            },
        )

        return ResolveResult.DROP

    def _check_implicit_drop_intention(self, object_full_name: str) -> bool:
        # Dropping any of parent objects implicitly drops the entire key
        if self.engine.intention_cache.check_parent_object_drop_intention(self.object_type, object_full_name):
            return True

        uk = self.existing_objects[object_full_name]
        table_full_name = f"{uk['database']}.{uk['schema']}.{uk['table']}"

        # Dropping one of unique key columns implicitly drops the entire key
        for col_name in uk["columns"]:
            if self.engine.intention_cache.check_column_drop_intention(table_full_name, col_name):
                return True

        return False



from snowddl.blueprint import SnapshotPolicyBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class SnapshotPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.SNAPSHOT_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW SNAPSHOT POLICIES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "schedule": r["schedule"],
                "expire_after_days": r["expire_after_days"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(SnapshotPolicyBlueprint)

    def create_object(self, bp: SnapshotPolicyBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE SNAPSHOT POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.schedule:
            query.append_nl(
                "SCHEDULE = {schedule}",
                {
                    "schedule": bp.schedule,
                },
            )

        if bp.expire_after_days:
            query.append_nl(
                "EXPIRE_AFTER_DAYS = {expire_after_days:d}",
                {
                    "expire_after_days": bp.expire_after_days,
                },
            )

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        self.engine.execute_unsafe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: SnapshotPolicyBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.schedule != row["schedule"]:
            if bp.schedule:
                self.engine.execute_unsafe_ddl(
                    "ALTER SNAPSHOT POLICY {full_name:i} SET SCHEDULE = {schedule}",
                    {
                        "full_name": bp.full_name,
                        "schedule": bp.schedule,
                    },
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER SNAPSHOT POLICY {full_name:i} UNSET SCHEDULE",
                    {
                        "full_name": bp.full_name,
                    },
                )

            result = ResolveResult.ALTER

        if bp.expire_after_days != row["expire_after_days"]:
            if bp.expire_after_days:
                self.engine.execute_unsafe_ddl(
                    "ALTER SNAPSHOT POLICY {full_name:i} SET EXPIRE_AFTER_DAYS = {expire_after_days:d}",
                    {
                        "full_name": bp.full_name,
                        "expire_after_days": bp.expire_after_days,
                    },
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER SNAPSHOT POLICY {full_name:i} UNSET EXPIRE_AFTER_DAYS",
                    {
                        "full_name": bp.full_name,
                    },
                )

            result = ResolveResult.ALTER

        if bp.comment != row["comment"]:
            self.engine.execute_unsafe_ddl(
                "ALTER SNAPSHOT POLICY {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
            )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP SNAPSHOT POLICY {database:i}.{schema:i}.{name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "name": row["name"],
            },
        )

        return ResolveResult.DROP



from functools import partial

from snowddl.blueprint import IcebergTableBlueprint, Ident, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.schema import schema_json_schema


# fmt: off
iceberg_table_json_schema = {
    "type": "object",
    "properties": {
        "catalog_table_name": {
            "type": "string"
        },
        "catalog_namespace": {
            "type": "string",
        },
        "metadata_file_path": {
            "type": "string"
        },
        "base_location": {
            "type": "string"
        },
        "replace_invalid_characters": {
            "type": "boolean"
        },
        "auto_refresh": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        },
    },
    "oneOf": [
        {"required": ["catalog_table_name"]},
        {"required": ["metadata_file_path"]},
        {"required": ["base_location"]},
    ],
    "additionalProperties": False
}
# fmt: on


class IcebergTableParser(AbstractParser):
    def load_blueprints(self):
        combined_params = {}

        for database_name in self.get_database_names():
            combined_params[database_name] = {}

            for schema_name in self.get_schema_names_in_database(database_name):
                schema_params = self.parse_single_entity_file(f"{database_name}/{schema_name}/params", schema_json_schema)
                combined_params[database_name][schema_name] = schema_params

        self.parse_schema_object_files(
            "iceberg_table", iceberg_table_json_schema, partial(self.process_table, combined_params=combined_params)
        )

    def process_table(self, f: ParsedFile, combined_params: dict):
        if not combined_params[f.database][f.schema].get("external_volume"):
            raise ValueError("Iceberg table requires parameter [external_volume] to be defined on schema level")

        if not combined_params[f.database][f.schema].get("catalog"):
            raise ValueError("Iceberg table requires parameter [catalog] to be defined on schema level")

        bp = IcebergTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            external_volume=Ident(combined_params[f.database][f.schema].get("external_volume")),
            catalog=Ident(combined_params[f.database][f.schema].get("catalog")),
            catalog_table_name=f.params.get("catalog_table_name"),
            catalog_namespace=f.params.get("catalog_namespace"),
            metadata_file_path=f.params.get("metadata_file_path"),
            base_location=f.params.get("base_location"),
            replace_invalid_characters=f.params.get("replace_invalid_characters", False),
            auto_refresh=f.params.get("auto_refresh", False),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from enum import Enum
from re import compile, IGNORECASE
from typing import List


class BaseDataType(Enum):
    NUMBER = {
        "base_name": "NUMBER",
        "number_of_properties": 2,
        "default_properties": [38, 0],
    }

    FILE = {
        "base_name": "FILE",
        "number_of_properties": 0,
        "default_properties": [],
    }

    FLOAT = {
        "base_name": "FLOAT",
        "number_of_properties": 0,
        "default_properties": [],
    }

    BINARY = {
        "base_name": "BINARY",
        "number_of_properties": 1,
        "default_properties": [8_388_608],
    }

    BOOLEAN = {
        "base_name": "BOOLEAN",
        "number_of_properties": 0,
        "default_properties": [],
    }

    VARCHAR = {
        "base_name": "VARCHAR",
        "number_of_properties": 1,
        "default_properties": [16_777_216],
    }

    DATE = {
        "base_name": "DATE",
        "number_of_properties": 0,
        "default_properties": [],
    }

    TIME = {
        "base_name": "TIME",
        "number_of_properties": 1,
        "default_properties": [9],
    }

    TIMESTAMP_LTZ = {
        "base_name": "TIMESTAMP_LTZ",
        "number_of_properties": 1,
        "default_properties": [9],
    }

    TIMESTAMP_NTZ = {
        "base_name": "TIMESTAMP_NTZ",
        "number_of_properties": 1,
        "default_properties": [9],
    }

    TIMESTAMP_TZ = {
        "base_name": "TIMESTAMP_TZ",
        "number_of_properties": 1,
        "default_properties": [9],
    }

    VARIANT = {
        "base_name": "VARIANT",
        "number_of_properties": 0,
        "default_properties": [],
    }

    OBJECT = {
        "base_name": "OBJECT",
        "number_of_properties": 0,
        "default_properties": [],
    }

    ARRAY = {
        "base_name": "ARRAY",
        "number_of_properties": 0,
        "default_properties": [],
    }

    GEOGRAPHY = {
        "base_name": "GEOGRAPHY",
        "number_of_properties": 0,
        "default_properties": [],
    }

    GEOMETRY = {
        "base_name": "GEOMETRY",
        "number_of_properties": 0,
        "default_properties": [],
    }

    VECTOR = {
        "base_name": "VECTOR",
        "number_of_properties": 2,
        "default_properties": [],
    }

    @property
    def number_of_properties(self) -> int:
        return self.value["number_of_properties"]

    @property
    def default_properties(self) -> List[int]:
        return self.value["default_properties"]

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}>"


class DataType:
    data_type_re = compile(
        r"^(?P<base_type>[a-z0-9_]+)(\((?P<val1>\d+)(,(?P<val2>\d+))?\)|\((?P<val1_vector>int|float),\s?(?P<val2_vector>\d+)\))?$",
        IGNORECASE,
    )

    def __init__(self, data_type_str):
        m = self.data_type_re.match(data_type_str)

        if not m:
            raise ValueError(f"Could not parse data type string [{data_type_str}]")

        try:
            self.base_type = BaseDataType[str(m["base_type"]).upper()]
        except KeyError:
            raise ValueError(f"Invalid base data type [{m['base_type']}] in type string [{data_type_str}]")

        if self.base_type == BaseDataType.VECTOR:
            self.val1 = str(m["val1_vector"]).upper()
            self.val2 = int(m["val2_vector"])
        else:
            self.val1 = int(m["val1"]) if self.base_type.number_of_properties >= 1 else None
            self.val2 = int(m["val2"]) if self.base_type.number_of_properties >= 2 else None

    @staticmethod
    def from_base_type(base_type: BaseDataType):
        if base_type.default_properties:
            return DataType(f"{base_type.name}({','.join(str(p) for p in base_type.default_properties)})")

        return DataType(base_type.name)

    def __str__(self):
        if self.base_type.number_of_properties >= 2:
            return f"{self.base_type.name}({self.val1},{self.val2})"
        elif self.base_type.number_of_properties >= 1:
            return f"{self.base_type.name}({self.val1})"

        return self.base_type.name

    def __repr__(self):
        return f"<{self.__class__.__name__}.{str(self)}>"

    def __eq__(self, other):
        if not isinstance(other, DataType):
            raise NotImplementedError

        return str(self) == str(other)



from snowddl.blueprint import WarehouseBlueprint
from snowddl.validator.abc_validator import AbstractValidator


class WarehouseValidator(AbstractValidator):
    def get_blueprints(self):
        return self.config.get_blueprints_by_type(WarehouseBlueprint)

    def validate_blueprint(self, bp: WarehouseBlueprint):
        self._validate_resource_constraint(bp)

    def _validate_resource_constraint(self, bp: WarehouseBlueprint):
        if bp.type == "SNOWPARK-OPTIMIZED" and bp.resource_constraint is None:
            raise ValueError(f"Resource constraint must be defined for SNOWPARK-OPTIMIZED warehouse [{bp.full_name}]")



from snowddl.blueprint import EventTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class EventTableResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.EVENT_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW EVENT TABLES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "owner": r["owner"],
                "retention_time": int(r["retention_time"]),
                "cluster_by": r["cluster_by"] if r["cluster_by"] else None,
                "change_tracking": bool(r["change_tracking"] == "ON"),
                "search_optimization": bool(r.get("search_optimization") == "ON"),
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(EventTableBlueprint)

    def create_object(self, bp: EventTableBlueprint):
        query = self._build_create_table(bp)
        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: EventTableBlueprint, row: dict):
        safe_alters = []
        unsafe_alters = []

        # Change tracking
        if bp.change_tracking != row["change_tracking"]:
            unsafe_alters.append(
                self.engine.format(
                    "SET CHANGE_TRACKING = {change_tracking:b}",
                    {
                        "change_tracking": bp.change_tracking,
                    },
                )
            )

        # Comment
        if bp.comment != row["comment"]:
            if bp.comment:
                safe_alters.append(
                    self.engine.format(
                        "SET COMMENT = {comment}",
                        {
                            "comment": bp.comment,
                        },
                    )
                )
            else:
                safe_alters.append(self.engine.format("UNSET COMMENT"))

        # Apply changes
        result = ResolveResult.NOCHANGE

        if safe_alters or unsafe_alters:
            for alter in safe_alters:
                self.engine.execute_safe_ddl(
                    "ALTER TABLE {full_name:i} {alter:r}",
                    {
                        "full_name": bp.full_name,
                        "alter": alter,
                    },
                )

            for alter in unsafe_alters:
                self.engine.execute_unsafe_ddl(
                    "ALTER TABLE {full_name:i} {alter:r}",
                    {
                        "full_name": bp.full_name,
                        "alter": alter,
                    },
                )

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl(
            "DROP TABLE {database:i}.{schema:i}.{table_name:i}",
            {
                "database": row["database"],
                "schema": row["schema"],
                "table_name": row["name"],
            },
        )

        return ResolveResult.DROP

    def _build_create_table(self, bp: EventTableBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE EVENT TABLE {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        if bp.change_tracking:
            query.append_nl("CHANGE_TRACKING = TRUE")

        if bp.comment:
            query.append_nl(
                "COMMENT = {comment}",
                {
                    "comment": bp.comment,
                },
            )

        return query



from snowddl.blueprint import (
    Ident,
    SemanticViewBlueprint,
    SemanticViewExpression,
    SemanticViewRelationship,
    SemanticViewTable,
    SchemaObjectIdent,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
semantic_view_json_schema = {
    "type": "object",
    "properties": {
        "tables": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "table_alias": {
                        "type": "string"
                    },
                    "table_name": {
                        "type": "string"
                    },
                    "primary_key": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "unique": {
                        "type": "array",
                        "items": {
                            "type": "array",
                            "items": {
                                "type": "string"
                            },
                            "minItems": 1
                        },
                        "minItems": 1
                    },
                    "with_synonyms": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "comment": {
                        "type": "string"
                    },
                },
                "required": ["table_name"],
            },
            "minItems": 1
        },
        "relationships": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "relationship_identifier": {
                        "type": "string"
                    },
                    "table_alias": {
                        "type": "string"
                    },
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "ref_table_alias": {
                        "type": "string"
                    },
                    "ref_columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                },
                "required": ["table_alias", "columns", "ref_table_alias", "ref_columns"],
            },
            "minItems": 1
        },
        "facts": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "table_alias": {
                        "type": "string"
                    },
                    "name": {
                        "type": "string"
                    },
                    "sql": {
                        "type": "string"
                    },
                    "with_synonyms": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "comment": {
                        "type": "string"
                    },
                },
                "required": ["table_alias", "name", "sql"],
            },
            "minItems": 1
        },
        "dimensions": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "table_alias": {
                        "type": "string"
                    },
                    "name": {
                        "type": "string"
                    },
                    "sql": {
                        "type": "string"
                    },
                    "with_synonyms": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "comment": {
                        "type": "string"
                    },
                },
                "required": ["table_alias", "name", "sql"],
            },
            "minItems": 1
        },
        "metrics": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "table_alias": {
                        "type": "string"
                    },
                    "name": {
                        "type": "string"
                    },
                    "sql": {
                        "type": "string"
                    },
                    "with_synonyms": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    },
                    "comment": {
                        "type": "string"
                    },
                },
                "required": ["table_alias", "name", "sql"],
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        },
    },
    "required": ["tables"],
    "additionalProperties": False
}
# fmt: on


class SemanticViewParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("semantic_view", semantic_view_json_schema, self.process_semantic_view)

    def process_semantic_view(self, f: ParsedFile):
        bp = SemanticViewBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            tables=[self.build_semantic_view_table(table_def, f) for table_def in f.params.get("tables", [])],
            relationships=[self.build_semantic_view_relationship(rel_def) for rel_def in f.params.get("relationships", [])],
            facts=[self.build_semantic_view_expression(expr_def) for expr_def in f.params.get("facts", [])],
            dimensions=[self.build_semantic_view_expression(expr_def) for expr_def in f.params.get("dimensions", [])],
            metrics=[self.build_semantic_view_expression(expr_def) for expr_def in f.params.get("metrics", [])],
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

    def build_semantic_view_table(self, table_def: dict, f: ParsedFile):
        return SemanticViewTable(
            table_alias=Ident(table_def.get("table_alias")),
            table_name=build_schema_object_ident(self.env_prefix, table_def.get("table_name"), f.database, f.schema),
            primary_key=[Ident(c) for c in table_def.get("primary_key")] if table_def.get("primary_key") else None,
            unique=[[Ident(c) for c in u] for u in table_def.get("unique")] if table_def.get("unique") else None,
            with_synonyms=[s for s in table_def.get("with_synonyms")] if table_def.get("with_synonyms") else None,
            comment=table_def.get("comment"),
        )

    def build_semantic_view_relationship(self, rel_def: dict):
        return SemanticViewRelationship(
            relationship_identifier=rel_def.get("relationship_identifier"),
            table_alias=Ident(rel_def.get("table_alias")),
            columns=[Ident(c) for c in rel_def.get("columns")],
            ref_table_alias=Ident(rel_def.get("ref_table_alias")),
            ref_columns=[Ident(c) for c in rel_def.get("ref_columns")],
        )

    def build_semantic_view_expression(self, expr_def: dict):
        return SemanticViewExpression(
            table_alias=Ident(expr_def.get("table_alias")),
            name=Ident(expr_def.get("name")),
            sql=self.normalise_sql_text_param(expr_def.get("sql")),
            with_synonyms=[s for s in expr_def.get("with_synonyms")] if expr_def.get("with_synonyms") else None,
            comment=expr_def.get("comment"),
        )



from snowddl.blueprint import (
    ViewBlueprint,
    ViewColumn,
    Ident,
    SchemaObjectIdent,
    ObjectType,
    AggregationPolicyBlueprint,
    AggregationPolicyReference,
    MaskingPolicyBlueprint,
    MaskingPolicyReference,
    ProjectionPolicyBlueprint,
    ProjectionPolicyReference,
    RowAccessPolicyBlueprint,
    RowAccessPolicyReference,
    build_schema_object_ident,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
view_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "additionalProperties": {
                "type": ["string", "null"]
            }
        },
        "text": {
            "type": "string"
        },
        "is_secure": {
            "type": "boolean"
        },
        "change_tracking": {
            "type": "boolean"
        },
        "depends_on": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        },
        "aggregation_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "required": ["policy_name"],
            "additionalProperties": False
        },
        "masking_policies": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "policy_name": {
                        "type": "string"
                    },
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "required": ["policy_name", "columns"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "projection_policies": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "policy_name": {
                        "type": "string"
                    },
                    "column": {
                        "type": "string",
                    }
                },
                "required": ["policy_name", "column"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "row_access_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                },
            },
            "required": ["policy_name", "columns"],
            "additionalProperties": False
        },
    },
    "required": ["text"],
    "additionalProperties": False
}
# fmt: on


class ViewParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("view", view_json_schema, self.process_view)

    def process_view(self, f: ParsedFile):
        column_blueprints = []

        for col_name, col_comment in f.params.get("columns", {}).items():
            column_blueprints.append(
                ViewColumn(
                    name=Ident(col_name),
                    comment=col_comment if col_comment else None,
                )
            )

        bp = ViewBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            text=self.normalise_sql_text_param(f.params["text"]),
            columns=column_blueprints if column_blueprints else None,
            is_secure=f.params.get("is_secure", False),
            change_tracking=f.params.get("change_tracking", False),
            depends_on=set(
                build_schema_object_ident(self.env_prefix, v, f.database, f.schema) for v in f.params.get("depends_on", [])
            ),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

        # Policies

        if f.params.get("aggregation_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["aggregation_policy"]["policy_name"], f.database, f.schema
            )

            ref = AggregationPolicyReference(
                object_type=ObjectType.VIEW,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["aggregation_policy"].get("columns", [])],
            )

            self.config.add_policy_reference(AggregationPolicyBlueprint, policy_name, ref)

        for mp in f.params.get("masking_policies", []):
            policy_name = build_schema_object_ident(self.env_prefix, mp["policy_name"], f.database, f.schema)

            ref = MaskingPolicyReference(
                object_type=ObjectType.VIEW,
                object_name=bp.full_name,
                columns=[Ident(c) for c in mp["columns"]],
            )

            self.config.add_policy_reference(MaskingPolicyBlueprint, policy_name, ref)

        for pp in f.params.get("projection_policies", []):
            policy_name = build_schema_object_ident(self.env_prefix, pp["policy_name"], f.database, f.schema)

            ref = ProjectionPolicyReference(
                object_type=ObjectType.VIEW,
                object_name=bp.full_name,
                column=Ident(pp["column"]),
            )

            self.config.add_policy_reference(ProjectionPolicyBlueprint, policy_name, ref)

        if f.params.get("row_access_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["row_access_policy"]["policy_name"], f.database, f.schema
            )

            ref = RowAccessPolicyReference(
                object_type=ObjectType.VIEW,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["row_access_policy"]["columns"]],
            )

            self.config.add_policy_reference(RowAccessPolicyBlueprint, policy_name, ref)



from snowddl.blueprint import AuthenticationPolicyBlueprint, AuthenticationPolicyReference, ObjectType, SchemaObjectIdent
from snowddl.error import SnowDDLExecuteError
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class AuthenticationPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.AUTHENTICATION_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta(
            "SHOW AUTHENTICATION POLICIES IN SCHEMA {database:i}.{schema:i}",
            {
                "database": schema["database"],
                "schema": schema["schema"],
            },
        )

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r["database_name"],
                "schema": r["schema_name"],
                "name": r["name"],
                "comment": r["comment"] if r["comment"] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(AuthenticationPolicyBlueprint)

    def create_object(self, bp: AuthenticationPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: AuthenticationPolicyBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._compare_policy(bp):
            result = ResolveResult.ALTER

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_policy_refs(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))
        self._drop_policy(SchemaObjectIdent("", row["database"], row["schema"], row["name"]))

        return ResolveResult.DROP

    def _create_policy(self, bp: AuthenticationPolicyBlueprint):
        query = self.engine.query_builder()

        query.append(
            "CREATE AUTHENTICATION POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        query.append_nl(
            "AUTHENTICATION_METHODS = ({authentication_methods})",
            {
                "authentication_methods": bp.authentication_methods,
            },
        )

        query.append_nl(
            "MFA_AUTHENTICATION_METHODS = ({mfa_authentication_methods})",
            {
                "mfa_authentication_methods": bp.mfa_authentication_methods,
            },
        )

        query.append_nl(
            "MFA_ENROLLMENT = {mfa_enrollment}",
            {
                "mfa_enrollment": bp.mfa_enrollment,
            },
        )

        query.append_nl(
            "CLIENT_TYPES = ({client_types})",
            {
                "client_types": bp.client_types,
            },
        )

        query.append_nl(
            "SECURITY_INTEGRATIONS = ({security_integrations})",
            {
                "security_integrations": bp.security_integrations,
            },
        )

        query.append_nl(
            "COMMENT = {comment}",
            {
                "comment": bp.comment,
            },
        )

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_authentication_policy)

    def _compare_policy(self, bp: AuthenticationPolicyBlueprint):
        applied_change = False

        cur = self.engine.execute_meta(
            "DESC AUTHENTICATION POLICY {full_name:i}",
            {
                "full_name": bp.full_name,
            },
        )

        row = {
            "authentication_methods": None,
            "mfa_authentication_methods": None,
            "mfa_enrollment": None,
            "client_types": None,
            "security_integrations": None,
            "comment": None,
        }

        for r in cur:
            k = r["property"].lower()

            if k not in row:
                continue

            if k == "comment" and r["value"] == "null":
                row[k] = None
            elif r["value"].startswith("["):
                row[k] = r["value"].strip("[]").split(", ")
            else:
                row[k] = r["value"]

        if bp.authentication_methods != row["authentication_methods"]:
            self.engine.execute_unsafe_ddl(
                "ALTER AUTHENTICATION POLICY {full_name:i} SET AUTHENTICATION_METHODS = ({authentication_methods})",
                {
                    "full_name": bp.full_name,
                    "authentication_methods": bp.authentication_methods,
                },
                condition=self.engine.settings.execute_authentication_policy,
            )

            applied_change = True

        if bp.mfa_authentication_methods != row["mfa_authentication_methods"]:
            self.engine.execute_unsafe_ddl(
                "ALTER AUTHENTICATION POLICY {full_name:i} SET MFA_AUTHENTICATION_METHODS = ({mfa_authentication_methods})",
                {
                    "full_name": bp.full_name,
                    "mfa_authentication_methods": bp.mfa_authentication_methods,
                },
                condition=self.engine.settings.execute_authentication_policy,
            )

            applied_change = True

        if bp.mfa_enrollment != row["mfa_enrollment"]:
            self.engine.execute_unsafe_ddl(
                "ALTER AUTHENTICATION POLICY {full_name:i} SET MFA_ENROLLMENT = {mfa_enrollment}",
                {
                    "full_name": bp.full_name,
                    "mfa_enrollment": bp.mfa_enrollment,
                },
                condition=self.engine.settings.execute_authentication_policy,
            )

            applied_change = True

        if bp.client_types != row["client_types"]:
            self.engine.execute_unsafe_ddl(
                "ALTER AUTHENTICATION POLICY {full_name:i} SET CLIENT_TYPES = ({client_types})",
                {
                    "full_name": bp.full_name,
                    "client_types": bp.client_types,
                },
                condition=self.engine.settings.execute_authentication_policy,
            )

            applied_change = True

        if bp.security_integrations != row["security_integrations"]:
            self.engine.execute_unsafe_ddl(
                "ALTER AUTHENTICATION POLICY {full_name:i} SET SECURITY_INTEGRATIONS = ({security_integrations:i})",
                {
                    "full_name": bp.full_name,
                    "security_integrations": bp.security_integrations,
                },
                condition=self.engine.settings.execute_authentication_policy,
            )

            applied_change = True

        if bp.comment != row["comment"]:
            self.engine.execute_unsafe_ddl(
                "ALTER AUTHENTICATION POLICY {full_name:i} SET COMMENT = {comment}",
                {
                    "full_name": bp.full_name,
                    "comment": bp.comment,
                },
                condition=self.engine.settings.execute_authentication_policy,
            )

            applied_change = True

        return applied_change

    def _drop_policy(self, policy: SchemaObjectIdent):
        self.engine.execute_unsafe_ddl(
            "DROP AUTHENTICATION POLICY {full_name:i}",
            {"full_name": policy},
            condition=self.engine.settings.execute_authentication_policy,
        )

    def _apply_policy_refs(self, bp: AuthenticationPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.full_name)
        applied_change = False

        for ref in bp.references:
            if ref.object_type == ObjectType.ACCOUNT:
                ref_key = ref.object_type.name
            else:
                ref_key = f"{ref.object_type.name}|{ref.object_name}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                del existing_policy_refs[ref_key]
                continue

            if ref.object_type == ObjectType.ACCOUNT:
                # Apply new policy for ACCOUNT
                if self._object_has_another_policy_ref(ref):
                    # FORCE is not supported, must unset existing policy first
                    self.engine.execute_unsafe_ddl(
                        "-- Previous policy must be removed before setting a new policy\n"
                        "ALTER ACCOUNT UNSET AUTHENTICATION POLICY",
                        condition=self.engine.settings.execute_authentication_policy
                        and self.engine.settings.execute_account_level_policy,
                    )

                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT SET AUTHENTICATION POLICY {policy_name:i}",
                    {
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_authentication_policy
                    and self.engine.settings.execute_account_level_policy,
                )
            else:
                # Apply new policy for USER (and other object types in future?)
                if self._object_has_another_policy_ref(ref):
                    # FORCE is not supported, must unset existing policy first
                    self.engine.execute_unsafe_ddl(
                        "-- Previous policy must be removed before setting a new policy\n"
                        "ALTER {object_type:r} {object_name:i} UNSET AUTHENTICATION POLICY",
                        {
                            "object_type": ref.object_type.name,
                            "object_name": ref.object_name,
                        },
                        condition=self.engine.settings.execute_authentication_policy,
                    )

                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} SET AUTHENTICATION POLICY {policy_name:i}",
                    {
                        "object_type": ref.object_type.name,
                        "object_name": ref.object_name,
                        "policy_name": bp.full_name,
                    },
                    condition=self.engine.settings.execute_authentication_policy,
                )

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            if existing_ref["object_type"] == ObjectType.ACCOUNT.name:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT UNSET AUTHENTICATION POLICY",
                    condition=self.engine.settings.execute_authentication_policy
                    and self.engine.settings.execute_account_level_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} UNSET AUTHENTICATION POLICY",
                    {
                        "object_type": existing_ref["object_type"],
                        "object_name": existing_ref["name"],
                    },
                    condition=self.engine.settings.execute_authentication_policy,
                )

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = self._get_existing_policy_refs(policy_name)

        for existing_ref in existing_policy_refs.values():
            if existing_ref["object_type"] == ObjectType.ACCOUNT.name:
                self.engine.execute_unsafe_ddl(
                    "ALTER ACCOUNT UNSET AUTHENTICATION POLICY",
                    condition=self.engine.settings.execute_authentication_policy
                    and self.engine.settings.execute_account_level_policy,
                )
            else:
                self.engine.execute_unsafe_ddl(
                    "ALTER {object_type:r} {object_name:i} UNSET AUTHENTICATION POLICY",
                    {
                        "object_type": existing_ref["object_type"],
                        "object_name": existing_ref["name"],
                    },
                    condition=self.engine.settings.execute_authentication_policy,
                )

    def _get_existing_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = {}

        cur = self.engine.execute_meta(
            "SELECT * FROM TABLE(snowflake.information_schema.policy_references(policy_name => {policy_name}))",
            {
                "policy_name": policy_name,
            },
        )

        for r in cur:
            if r["REF_ENTITY_DOMAIN"] == ObjectType.ACCOUNT.name:
                ref_key = r["REF_ENTITY_DOMAIN"]
            else:
                ref_key = f"{r['REF_ENTITY_DOMAIN']}|{r['REF_ENTITY_NAME']}"

            existing_policy_refs[ref_key] = {
                "object_type": r["REF_ENTITY_DOMAIN"],
                "name": r["REF_ENTITY_NAME"],
            }

        return existing_policy_refs

    def _object_has_another_policy_ref(self, ref: AuthenticationPolicyReference):
        try:
            cur = self.engine.execute_meta(
                "SELECT * FROM TABLE(snowflake.information_schema.policy_references(ref_entity_domain => {object_type}, ref_entity_name  => {object_name}))",
                {
                    "object_type": ref.object_type.name,
                    "object_name": self.engine.context.current_account
                    if ref.object_type == ObjectType.ACCOUNT
                    else ref.object_name,
                },
            )
        except SnowDDLExecuteError as e:
            # User does not exist or not authorized
            # Skip this error during planning
            if e.snow_exc.errno == 2003:
                return False
            else:
                raise

        for r in cur:
            if r["POLICY_KIND"] == "AUTHENTICATION_POLICY":
                return True

        return False



from fnmatch import translate
from re import compile
from string import ascii_letters, digits
from typing import List, Pattern

from snowddl.blueprint.ident import AbstractIdentWithPrefix


class IdentPattern:
    allowed_chars_pattern = set(ascii_letters + digits + "_$()." + "|!*?[]")
    special_chars_complex_pattern = set("|!*?[]")

    def __init__(self, pattern):
        self.pattern = self._validate_pattern(pattern)
        self.is_complex_pattern = self._is_complex_pattern(pattern)

        self.include_regexp: List[Pattern] = []
        self.exclude_regexp: List[Pattern] = []

        if self.is_complex_pattern:
            for sub_pattern in self.pattern.split("|"):
                is_exclude = False

                if sub_pattern.startswith("!"):
                    is_exclude = True
                    sub_pattern = sub_pattern[1:]

                if not sub_pattern:
                    continue

                compiled_regexp = compile(translate(sub_pattern))

                if is_exclude:
                    self.exclude_regexp.append(compiled_regexp)
                else:
                    self.include_regexp.append(compiled_regexp)

            if not self.include_regexp:
                raise ValueError(f"Identifier pattern [{self.pattern}] does not contain any positive sub-patterns")

    @classmethod
    def build_from_ident(cls, ident: AbstractIdentWithPrefix):
        return cls(cls._get_str_ident_without_prefix(ident))

    def is_match_ident(self, ident: AbstractIdentWithPrefix) -> bool:
        str_ident_without_prefix = self._get_str_ident_without_prefix(ident)

        if self.is_complex_pattern:
            is_match_include = any(regexp.match(str_ident_without_prefix) for regexp in self.include_regexp)
            is_match_exclude = any(regexp.match(str_ident_without_prefix) for regexp in self.exclude_regexp)

            return is_match_include and not is_match_exclude

        return str_ident_without_prefix == self.pattern

    def __str__(self):
        return self.pattern

    def __repr__(self):
        return f"<{self.__class__.__name__}={str(self)}>"

    def _is_complex_pattern(self, val):
        return any(char in self.special_chars_complex_pattern for char in val)

    @classmethod
    def _get_str_ident_without_prefix(cls, ident: AbstractIdentWithPrefix):
        return str(ident).removeprefix(ident.env_prefix)

    def _validate_pattern(self, val):
        val = str(val)

        if not val:
            raise ValueError("Pattern cannot be empty")

        for char in val:
            if char not in self.allowed_chars_pattern:
                raise ValueError(
                    f"Character [{char}] is not allowed in identifier pattern [{val}], "
                    f"only ASCII letters, digits, single underscores and Unix-style pattern characters are accepted"
                )

        return val.upper()



from snowddl.blueprint import SequenceBlueprint, SchemaObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


# fmt: off
sequence_json_schema = {
    "type": "object",
    "properties": {
        "start": {
            "type": "integer"
        },
        "interval": {
            "type": "integer"
        },
        "is_ordered": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}
# fmt: on


class SequenceParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("sequence", sequence_json_schema, self.process_sequence)

    def process_sequence(self, f: ParsedFile):
        bp = SequenceBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            start=f.params.get("start", 1),
            interval=f.params.get("interval", 1),
            is_ordered=f.params.get("is_ordered"),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)



from functools import partial

from snowddl.blueprint import (
    AccountObjectIdent,
    DynamicTableBlueprint,
    SchemaObjectIdent,
    DynamicTableColumn,
    Ident,
    ObjectType,
    build_schema_object_ident,
    AggregationPolicyBlueprint,
    AggregationPolicyReference,
    MaskingPolicyBlueprint,
    MaskingPolicyReference,
    ProjectionPolicyBlueprint,
    ProjectionPolicyReference,
    RowAccessPolicyBlueprint,
    RowAccessPolicyReference,
)
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.schema import database_json_schema, schema_json_schema


# fmt: off
dynamic_table_json_schema = {
    "type": "object",
    "properties": {
        "columns": {
            "type": "object",
            "additionalProperties": {
                "type": ["string", "null"]
            }
        },
        "text": {
            "type": "string"
        },
        "target_lag": {
            "type": "string"
        },
        "warehouse": {
            "type": "string",
        },
        "refresh_mode": {
            "type": "string",
        },
        "initialize": {
            "type": "string",
        },
        "cluster_by": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "depends_on": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        },
        "aggregation_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                }
            },
            "required": ["policy_name"],
            "additionalProperties": False
        },
        "masking_policies": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "policy_name": {
                        "type": "string"
                    },
                    "columns": {
                        "type": "array",
                        "items": {
                            "type": "string"
                        },
                        "minItems": 1
                    }
                },
                "required": ["policy_name", "columns"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "projection_policies": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "policy_name": {
                        "type": "string"
                    },
                    "column": {
                        "type": "string",
                    }
                },
                "required": ["policy_name", "column"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "row_access_policy": {
            "type": "object",
            "properties": {
                "policy_name": {
                    "type": "string"
                },
                "columns": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "minItems": 1
                },
            },
            "required": ["policy_name", "columns"],
            "additionalProperties": False
        },
    },
    "required": ["text", "target_lag", "warehouse"],
    "additionalProperties": False
}
# fmt: on


class DynamicTableParser(AbstractParser):
    unit_singular_to_plural = {
        "second": "seconds",
        "minute": "minutes",
        "hour": "hours",
        "day": "days",
    }

    unit_plural_to_singular = {v: k for k, v in unit_singular_to_plural.items()}

    def load_blueprints(self):
        combined_params = {}

        for database_name in self.get_database_names():
            database_params = self.parse_single_entity_file(f"{database_name}/params", database_json_schema)
            combined_params[database_name] = {}

            for schema_name in self.get_schema_names_in_database(database_name):
                schema_params = self.parse_single_entity_file(f"{database_name}/{schema_name}/params", schema_json_schema)

                combined_params[database_name][schema_name] = {
                    "is_transient": database_params.get("is_transient", False) or schema_params.get("is_transient", False),
                    "retention_time": schema_params.get("retention_time"),
                }

        self.parse_schema_object_files(
            "dynamic_table", dynamic_table_json_schema, partial(self.process_dynamic_table, combined_params=combined_params)
        )

    def process_dynamic_table(self, f: ParsedFile, combined_params: dict):
        column_blueprints = []

        for col_name, col_comment in f.params.get("columns", {}).items():
            column_blueprints.append(
                DynamicTableColumn(
                    name=Ident(col_name),
                    comment=col_comment if col_comment else None,
                )
            )

        bp = DynamicTableBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            text=self.normalise_sql_text_param(f.params["text"]),
            columns=column_blueprints if column_blueprints else None,
            target_lag=self.normalise_target_lag(f.params["target_lag"]),
            warehouse=AccountObjectIdent(self.env_prefix, f.params["warehouse"]),
            refresh_mode=f.params.get("refresh_mode").upper() if f.params.get("refresh_mode") else None,
            initialize=f.params.get("initialize").upper() if f.params.get("initialize") else None,
            cluster_by=f.params.get("cluster_by"),
            is_transient=f.params.get("is_transient", combined_params[f.database][f.schema].get("is_transient", False)),
            retention_time=f.params.get("retention_time", combined_params[f.database][f.schema].get("retention_time", None)),
            depends_on=set(
                build_schema_object_ident(self.env_prefix, d, f.database, f.schema) for d in f.params.get("depends_on", [])
            ),
            comment=f.params.get("comment"),
        )

        self.config.add_blueprint(bp)

        # Policies

        if f.params.get("aggregation_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["aggregation_policy"]["policy_name"], f.database, f.schema
            )

            ref = AggregationPolicyReference(
                object_type=ObjectType.DYNAMIC_TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["aggregation_policy"].get("columns", [])],
            )

            self.config.add_policy_reference(AggregationPolicyBlueprint, policy_name, ref)

        for mp in f.params.get("masking_policies", []):
            policy_name = build_schema_object_ident(self.env_prefix, mp["policy_name"], f.database, f.schema)

            ref = MaskingPolicyReference(
                object_type=ObjectType.DYNAMIC_TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in mp["columns"]],
            )

            self.config.add_policy_reference(MaskingPolicyBlueprint, policy_name, ref)

        for pp in f.params.get("projection_policies", []):
            policy_name = build_schema_object_ident(self.env_prefix, pp["policy_name"], f.database, f.schema)

            ref = ProjectionPolicyReference(
                object_type=ObjectType.DYNAMIC_TABLE,
                object_name=bp.full_name,
                column=Ident(pp["column"]),
            )

            self.config.add_policy_reference(ProjectionPolicyBlueprint, policy_name, ref)

        if f.params.get("row_access_policy"):
            policy_name = build_schema_object_ident(
                self.env_prefix, f.params["row_access_policy"]["policy_name"], f.database, f.schema
            )

            ref = RowAccessPolicyReference(
                object_type=ObjectType.DYNAMIC_TABLE,
                object_name=bp.full_name,
                columns=[Ident(c) for c in f.params["row_access_policy"]["columns"]],
            )

            self.config.add_policy_reference(RowAccessPolicyBlueprint, policy_name, ref)

    def normalise_target_lag(self, target_lag: str):
        if target_lag.upper() == "DOWNSTREAM":
            return "DOWNSTREAM"

        num, _, unit = target_lag.partition(" ")

        num = int(num)
        unit = unit.lower()

        if num > 1 and (unit in self.unit_singular_to_plural):
            unit = self.unit_singular_to_plural[unit]
        elif num == 1 and (unit in self.unit_plural_to_singular):
            unit = self.unit_plural_to_singular[unit]

        return f"{num} {unit}"
